#!WPY

# tommy's attempt at writing an AI for the Battle for Wesnoth

import wesnoth
import math
import random
#import time
# would like to add true division, but this is unsupported by
# the current restricted python framework used.
# if modifying, be careful about integer division. it's a b*tch to debug.
#from __future__ import division 


# DEBUGLEVEL can be 0, 1, 2, 3 or 4.
# 0 means no output, 1 means serious errors only, 2 adds warnings and some
# general info, 3 adds debug info, 4 adds very detailed debug
# info. Debug level 4 causes A LOT of output. YOU WERE WARNED.
DEBUGLEVEL = 2

# if TTYDEBUG is set to true, debug output is primted to the terminal
# wesnoth was run from. Set it to false to have output displayed through
# the normal method (chat messages if wesnoth is started in debug mode)
TTYDEBUG = True

def displaystring(string):
    '''
    display a message, either in the terminal wesnoth was run from,
    or through the internal debugging view
    '''
    if TTYDEBUG:
        print string
    else:
        wesnoth.log_message(string)

def megadprint(string):
    '''
    print mega-detailed debug messages if debug level >= 4
    '''
    if DEBUGLEVEL > 3:
        displaystring(string)

def dprint(string):
    '''
    print a debugging message if debug level >= 3
    '''
    if DEBUGLEVEL > 2:
        displaystring(string)

def wprint(string):
    '''
    print a warning message, if debug level >= 2
    '''
    if DEBUGLEVEL > 1:
        displaystring(string)

def eprint(string):
    '''
    print an error message, if debug level is >= 1
    '''
    if DEBUGLEVEL > 0:
        displaystring(string)

def iprint(string):
    '''
    print an informational message, if debug level is >= 2
    '''
    if DEBUGLEVEL > 1:
        displaystring(string)

def printable(location): # convenience function for debug output
    x = str(location.x + 1)
    y = str(location.y + 1)
    return '(' + x + ", " + y + ')'

class TommyAI(object):
    '''
    A python AI, written by Tommy
    '''
    def __init__(self):
        '''
        Do AI stuff, kazablamm!
        '''
        self.myteam = wesnoth.get_current_team()
        self.gamemap = wesnoth.get_map()
        self.gamestatus = wesnoth.get_gamestatus()
        
        dprint("  +===============================================+")
        dprint("  | Tommy's Python AI performing moves for team %s |" % \
            self.myteam.side)
        dprint("  +===============================================+")
        
        self.setallunits()
        self.castletiles = []
        self.keeps = {}
        self.villages = []
        self.criticallocations = {}
        self.attackedlocations = []
        self.defensescores = {}
        self.movementscores = {}
        self.combatscores = {}
        self.myrecruittypes = self.myteam.recruits()
        self.losing = False
        self.winning = False
        self.attackcache = {}
        self.orders = {}
        self.recruitedthisturn = False
        self.myteamlawfulness = 0
        self.contactmade = False
        self.villagedibs = {}
        
        self.loadvariables()
        # currently this is only run once per turn:
        self.analysegamemap()
        
        acted = True
        while acted:  # loop until no units act
            acted = False
            self.myunits = self.getmyunits()
            self.myunits.sort(cmp=self.sortcompare, reverse=True)
            for index in xrange(len(self.myunits)):
                location, unit, name = self.myunits[index]
                wesnoth.raise_user_interact()
                self.setallunits(mylocation=location)
                if unit.can_recruit:
                    unitacted, resetturn = self.doleader(location, unit, name)
                else:
                    unitacted, resetturn = self.dounit(location, unit, name)
                if resetturn:
                    acted = True
                    break
                acted = acted or unitacted
        
        self.savevariables()
        
        dprint("  +----------------------------------------------+")
        dprint("  | Tommy's Python AI has finished moving team %s |" % \
            self.myteam.side)
        dprint("  +----------------------------------------------+")
    
    def analysegamemap(self):
        '''
        Analyse the map and create the following data structures:
            self.castletiles = list of castle locations
            self.keeps = dictionary of keep locations -> number of
                attached castle tiles
            self.villages = list of village locations
            self.defensescores[unitname] = average chance to be hit
                over entire map for each recruitable unit type
            self.movementscores[unitname] = average movement value
                over entire map for each recruitable unit type
            self.criticallocations = dictionary of locations -> weights
                -> villages (weight 8.0 -> 0.25 depending on owner)
                -> any tile with average defense >= 50% out of all unit types
                   (weight average defense)
                
        '''
        impassabletiles = {}
        self.criticallocations = {}
        for recruittype in self.myrecruittypes:
            self.defensescores[recruittype.name] = 0.0
            self.movementscores[recruittype.name] = 0.0
            impassabletiles[recruittype.name] = 0.0
        alltypes = []
        totalworth = 0.0
        for location, unit in wesnoth.get_units().iteritems():
            type = unit.type()
            totalworth += self.getworth(unit)
            if not type.name in self.defensescores:
                alltypes.append(type)
                self.defensescores[type.name] = 0.0
                self.movementscores[type.name] = 0.0
                impassabletiles[type.name] = 0.0
        self.averageworth = totalworth / len(alltypes) # please never be 0
        teams = wesnoth.get_teams()
        movescores = {} # unit -> {move cost -> number of applicable tiles}
        defensescores = {} # unit -> {defense -> number of applicable tiles}
        totalallyvillages = 0
        for x in range(self.gamemap.x):
            for y in range(self.gamemap.y):
                location = wesnoth.get_location(x, y)
                alreadycritical = False
                if self.gamemap.is_castle(location):
                    self.castletiles.append(location)
                if self.gamemap.is_keep(location):
                    self.keeps[location] = []
                if self.gamemap.is_village(location):
                    self.villages.append(location)
                    for team in teams:
                        if team.owns_village(location):
                            if team.is_enemy:
                                # low priority, but still denying
                                # them access to their villages might
                                # be worth something
                                self.criticallocations[location] = 2.0
                                alreadycritical = True
                                break
                            elif team.side == self.myteam.side:
                                # high priority, assume if a village
                                # is captured we won't get it back for
                                # 5 turns... might need tweaking
                                self.criticallocations[location] = 20.0
                                alreadycritical = True
                                totalallyvillages += 1
                                break
                            else: # ally owns
                                totalallyvillages += 1
                                break
                    if not alreadycritical:
                        # unowned or ally owns village
                        self.criticallocations[location] = 8.0
                        alreadycritical = True
                totalctbh = 0.0
                typescounted = 0.0
                for type in (self.myrecruittypes + alltypes):
                    movementcost = type.movement_cost(location)
                    if movementcost > type.movement:
                        impassabletiles[type.name] += 1.0
                    else:
                        if type.name not in movescores:
                            movescores[type.name] = {}
                        if type.name not in defensescores:
                            defensescores[type.name] = {}
                        averagemove = type.movement / float(movementcost)
                        if averagemove not in movescores[type.name]:
                            movescores[type.name][averagemove] = 1
                        else:
                            movescores[type.name][averagemove] += 1
                        chancetobehit = \
                            type.defense_modifier(location) / 100.0
                        if chancetobehit not in defensescores[type.name]:
                            defensescores[type.name][chancetobehit] = 1
                        else:
                            defensescores[type.name][chancetobehit] += 1
                        totalctbh += chancetobehit
                        typescounted += 1.0
                if typescounted and totalctbh:
                    averagedefense = 1.0 - (totalctbh / typescounted)
                    if averagedefense >= 0.55:
                        # assume defensible positions are worth around
                        # 2 * defense in gold
                        averagedefense *= 2.0
                        if alreadycritical:
                            self.criticallocations[location] += averagedefense
                        else:
                            self.criticallocations[location] = averagedefense
        if 2 * totalallyvillages < len(self.villages):
            # having at least 50% of the villages on the map is crucial.
            self.losing = True
        for name in movescores:
            # want to ignore the worst 2/3 of the map for each unit
            # they can just avoid it
            total = 0.0
            sortablescores = []
            for score, number in movescores[name].iteritems():
                total += number
                sortablescores.append((score, number))
            topthird = total / 3.0
            totalscore = 0.0
            totalnumber = 0.0
            sortablescores.sort(reverse=True)
            for score, number in sortablescores:
                totalscore += score * number
                totalnumber += number
                topthird -= number
                if topthird <= 0:
                    break
            average = totalscore / float(totalnumber)
            self.movementscores[name] = average
            megadprint("setting move score for %s to %s" % (name, average))
        for name in defensescores:
            # ignore the worst 2/3, unit can choose not to be there
            total = 0.0
            sortablescores = []
            for score, number in defensescores[name].iteritems():
                total += number
                sortablescores.append((score, number))
            topthird = total / 3.0
            totalscore = 0.0
            totalnumber = 0.0
            sortablescores.sort()
            for score, number in sortablescores:
                totalscore += score * number
                totalnumber += number
                topthird -= number
                if topthird <= 0:
                    break
            average = totalscore / float(totalnumber)
            self.defensescores[name] = average
            megadprint("setting defense score for %s to %s" % (name, average))
        for keeptile in self.keeps:
            checkedtiles = [keeptile]
            def castlewalk(startingtile):
                surroundings = wesnoth.get_adjacent_tiles(startingtile)
                for tile in surroundings:
                    if (self.gamemap.is_castle(tile) \
                    or self.gamemap.is_keep(tile)):
                        if tile not in checkedtiles:
                            checkedtiles.append(tile)
                            castlewalk(tile)
            castlewalk(keeptile)
            checkedtiles.remove(keeptile)
            self.keeps[keeptile] = checkedtiles
            
        numtiles = self.gamemap.x * self.gamemap.y
        dprint("::: Successfully analysed gamemap :::")
        dprint("%s critical locations found out of %s total map tiles" % \
            (len(self.criticallocations), numtiles) + \
            ", of which %s are villages" % len(self.villages))
    
    def getmyunits(self):
        '''
        Return a list of units belonging to the current team.
        List entries are tuples of the format:
            (location, unit, unittype.name)
        '''
        allunits = wesnoth.get_units()
        myunits = []
        lawfulness = 0
        for location, unit in allunits.iteritems():
            if unit.side == self.myteam.side:
                unitname = self.getname(unit)
                myunits.append((location, unit, unitname))
                lawfulness -= (unit.type().alignment - 1)
        self.myteamlawfulness = lawfulness / float(len(myunits))
        return myunits
    
    def getname(self, unit):
        '''
        Returns the readable name of the unit, mainly for debug output
        '''
        unitname = ""
        #if unit.can_recruit:
        #    unitname = "Leader "
        if unit.name:
            if unit.type().name in unit.name:
                unitname += unit.name
            else:
                unitname += "%s (%s)" % (unit.name, unit.type().name)
        else:
            unitname += unit.type().name
        return unitname
    
    def getworth(self, unit):
        '''
        Return how much a certain unit is worth.
        
        Currently us the base cost of the unit plus the gold worth of its xp.
        '''
        unittype = unit.type()
        if unittype.advances_to():
            worth = unittype.cost + self.xptogold(unit.experience)
        else:
            worth = unittype.cost
        if unit.can_recruit:
            # this value may need tweaking. Or should be set by WML?
            worth += 20
        return worth
    
    def xptogold(self, xp):
        '''
        Return the equivalent worth in gold of 'xp' experience points.
        
        Currently this is just 1.5 gold for every 1 xp.
        '''
        return xp * 1.5
    
    def hptogold(self, hp, unit):
        '''
        Returns the effective worth of 'hp' damage or healing for 'unit'.
        
        Simply uses self.getworth(unit) to determine what the unit is
        worth, and works out what percentage of its max hp applies.
        '''
        unitworth = self.getworth(unit)
        return unitworth * (hp / float(unit.max_hitpoints))
    
    def xpfromunit(self, unit):
        '''
        Returns the amount of xp killing a unit is worth
        '''
        return (unit.type().level * 8) or 4
    
    def sortcompare(self, x, y):
        '''
        Helper function to sort units by speed, hp and xp to level.
        Always puts the leader first.
        This way:
            * scouts move last, giving other units a chance to capture
                the nearest villages
            * units that can level get the unit's weakened from attacking
            * wounded units can more reliably hide behind other units
            * units get recruited at the end of the turn not the beginning
        Some issues:
            * healthier units steal villages
            * in a small corridor, slightly faster units cannot overtake
        '''
        # higher values mean move sooner
        unitx = x[1]
        unity = y[1]
        # always move the leader first
        if unitx.can_recruit or unity.can_recruit:
            return cmp(unitx.can_recruit, unity.can_recruit)
        # for now I'm going to always put backstabbing units last.
        # this should be changed when the leader gains some idea of
        # how to use them.
        xbackstabs = None
        ybackstabs = None
        for attacktype in unitx.attacks():
            if attacktype.backstab:
                xbackstabs = True
        for attacktype in unity.attacks():
            if attacktype.backstab:
                ybackstabs = True
        if xbackstabs or ybackstabs:
            return cmp(ybackstabs, xbackstabs) # note reversed
        hpx = unitx.hitpoints * 1.0 / unitx.max_hitpoints
        hpy = unity.hitpoints * 1.0 / unity.max_hitpoints
        xpx = unitx.experience * 1.0 / unitx.max_experience
        xpy = unity.experience * 1.0 / unity.max_experience
        movex = unitx.max_movement / 10.0
        movey = unity.max_movement / 10.0
        mixx = hpx + xpx - movex
        mixy = hpy + xpy - movey
        return cmp(mixx, mixy)
    
    def setallunits(self, mylocation=None):
        '''
        Return a dictionary of location -> unit items.
        
        Removes stoned units, and the current units.
        '''
        self.allunits = wesnoth.get_units()
        if mylocation:
            # don't count yerself, ye might move, fool
            if mylocation not in self.allunits:
                eprint("BLEGH can't remove %s, not in allunits" % \
                    printable(mylocation))
            self.allunits.pop(mylocation)
        self.allenemies = {}
        for unitlocation in self.allunits.keys():
            unit = self.allunits[unitlocation]
            if unit.stoned:
                # it's ok baby, it's just a statue, it's NOT REAL
                self.allunits.pop(unitlocation)
            elif unit.is_enemy:
                self.allenemies[unitlocation] = unit
    
    def savevariables(self):
        '''
        Save anything the AI should remember next turn
        '''
        def setvariable(name, string):
            '''
            Set a variable and print a debugging message.
            '''
            wesnoth.set_variable(name, string)
            dprint("Storing variable: '%s' = '%s'" % (name, string))
        if self.hatedunits:
            hatedunits = []
            hatred = []
            # store high value targets and their values
            for name, value in self.hatedunits.iteritems():
                hatedunitslist.append(name)
                hatredlist.append(str(value))
            setvariable('hatedunits', ','.join(hatedunits))
            setvariable('hatred', ','.join(hatred))
        if self.contactmade:
            setvariable('contactmade', 'True')
        if self.villagedibs:
            dibs = []
            for location, unitname in self.villagedibs.iteritems():
                villagestring = "(%s,%s)" % (location.x, location.y)
                dibsstring = villagestring + ':' + unitname
                dibs.append(dibsstring)
            setvariable('villagedibs', ';'.join(dibs))
    
    def loadvariables(self):
        '''
        Interperet any variables saved last turn
        '''
        hatedunitsstring = wesnoth.get_variable('hatedunits')
        hatredstring = wesnoth.get_variable('hatred')
        if hatedunitsstring:
            hatedunits = hatedunitsstring.split(',')
            hated = hatedstring.split(',')
            for index in xrange(len(hatedunits)):
                self.hatedunits[hatedunits[index]] = hatred[index]
            dprint("Loaded variable: 'hatedunits'")
        else:
            self.hatedunits = {}
        contactmade = wesnoth.get_variable('contactmade')
        if contactmade == 'True':
            dprint("Loaded variable: 'contactmade' = 'True'")
            self.contactmade = True
        villagedibs = wesnoth.get_variable('villagedibs')
        if villagedibs:
            dibs = villagedibs.split(';')
            for dibsstring in dibs:
                locationstring, unitname = dibsstring.split(':')
                x, y = locationstring[1:-1].split(',')
                x, y = int(x), int(y)
                location = wesnoth.get_location(x, y)
                if self.gamemap.is_village(location):
                    allunits = wesnoth.get_units()
                    visibleenemies = wesnoth.get_enemy_destinations_by_unit()
                    if location in allunits:
                        if not allunits[location].is_enemy:
                            # ally got it?
                            continue
                        elif location in visibleenemies:
                            # enemy got it
                            continue
                        else:
                            # someone got it but we shouldn't know they did,
                            # so pretend we didn't see them.
                            pass
                    dprint("Loading village dibs '%s'" % dibsstring)
                    self.villagedibs[location] = unitname
    
    def getleaders(self):
        '''
        Return a list of (location, unit) pairs pointing to all leaders.
        
        Includes leaders who are in shroud/fog or otherwise hidden.
        '''
        leaders = []
        for location, unit in self.allunits.iteritems():
            if unit.can_recruit:
                leaders.append((location, unit))
        return leaders
    
    def getdistance(self, unit, source, destination):
        '''
        Return some vague value representing the distance from source to
        destination for unit.
        
        Currently... this is just the distance in tiles, normalized by the
        unit's average movement on this map.
        '''
        if source == destination:
            # trivial case
            return 0
        basedistance = float(source.distance_to(destination))
        distance = basedistance / self.movementscores[unit.type().name]
        return distance
    
    def getopposite(self, tile1, tile2):
        '''
        Return the tile on the opposite side of tile2 from tile1, or None
        if is outside the bounds of the map.
        '''
        x1, y1 = tile1.x, tile1.y
        x2, y2 = tile2.x, tile2.y
        x3 = x2 + (x2 - x1)
        # too lazy to figure out how to do this in one line
        if y1 != y2:
            y3 = y2
        else:
            if x2 & 1: # x2 is even
                y3 = y1 + 1
            else:
                y3 = y1 - 1
        return wesnoth.get_location(x3, y3)
    
    def hitchance(self, attacktype, basehitchance):
        '''
        Returns the chance to hit of an attack after factoring in specials
        such as magic and marksman
        '''
        hitchance = basehitchance
        if attacktype.magical:
            hitchance = 0.7
        if attacktype.marksman:
            hitchance = max(0.6, hitchance)
        return hitchance
    
    def bestretaliation(self, attackertype, defendertype, attackerctbh, \
    attacktype, nextturn=False, side=0, generic=False, lawfulbonus=0):
        '''
        Return the index of the best retaliation to a specific attacktype
        (or None if the unit can not retaliate) and the self.attackscore of
        the retaliation.
        
        Currently simply uses attackscore to decide which is best.
        '''
        bestindex = None
        bestscore = 0.0
        maxscore = 0.0
        added = 0.0
        prevented = 0.0
        retaliations = defendertype.attacks()
        for retaliation in retaliations:
            index = retaliations.index(retaliation)
            if retaliation.range == attacktype.range:
                retaliationstats = self.attackscore( \
                    defendertype, attackertype, attackerctbh, index, \
                    nextturn=nextturn, side=side, retaliation=True, \
                    generic=generic, lawfulbonus=lawfulbonus)
                if retaliationstats[0] > bestscore:
                    bestscore = retaliationstats[0]
                    bestindex = index
                    maxscore = retaliationstats[1]
                    if generic:
                        added = retaliationstats[2]
                        prevented = retaliationstats[3]
        if not generic:
            return bestindex, bestscore, maxscore
        else:
            return bestindex, bestscore, maxscore, added, prevented
    
    def attackscore(self, attackertype, defendertype, hitchance, attackindex, \
    nextturn=False, side=0, retaliation=False, generic=False, lawfulbonus=0):
        '''
        Estimates the value of an attack vs. a certain foe, with the given
        chance to hit.
        The basic formula is:
          num_attacks * damage * damagefrom(attacktype) * chancetohit
        Most attack specials are not considered, as their worth depends on
        the status of the specific units, only steadfast is counted here.
        
        Effectively the attackscore is the expected damage dealt, in hp.
        
        Also returns the maximum damage possibly dealt.
        '''
        # I think this will break if the unit's attacks have changed
        # from the default for its type...
        attacktype = attackertype.attacks()[attackindex]
        score = attacktype.num_attacks * attacktype.damage
        damagefromtype = defendertype.damage_from(attacktype) / 100.0
        if not retaliation and defendertype.steadfast and damagefromtype < 1.0:
            damagefromtype = damagefromtype ** 2
        score *= damagefromtype
        maxscore = score
        score *= self.hitchance(attacktype, hitchance)
        if not generic:
            score *= self.todmodifier(attackertype, side, nextturn)
            maxscore *= self.todmodifier(attackertype, side, nextturn)
            return score, maxscore
        
        added = 0
        prevented = 0
        if generic:
            # estimate modifiers for specials
            score *= 1.0 - (lawfulbonus * attackertype.alignment)
            basescore = score
            maxscore *= 1.0 - (lawfulbonus * attackertype.alignment)
            if attacktype.drains and not defendertype.not_living:
                prevented = score * 0.4
            if attacktype.poison and not defendertype.not_living:
                if defendertype.regenerate:
                    added = 8
                else:
                    added = 16
            if attacktype.slow:
                prevented += self.valueofslow(attacktype, defendertype, \
                    retaliation=retaliation)
            if defendertype.regenerate:
                added -= 4 # or so
        return basescore, maxscore, added, prevented
    
    def valueofslow(self, attacktype, defendertype, retaliation=False):
        '''
        Given an attacktype with the 'slow' special, returns the quivalent
        amount of damage slowing this unit is worth.
        
        Currently is worth 1/4 of their max damage in the opposite
        range to the attack, or if our slowing attack is a retaliation,
        1/4 of the max damage in the same range.
        '''
        if not attacktype.slow:
            return
        if attacktype.range == 'ranged':
            if retaliation:
                range = 'ranged'
            else:
                range = 'melee'
        else:
            if retaliation:
                range = 'melee'
            else:
                range = 'ranged'
        highestdamage = 0.0
        for attacktype in defendertype.attacks():
            if attacktype.range == range:
                averagedamage = \
                    attacktype.damage * attacktype.num_attacks * 0.5
                if averagedamage > highestdamage:
                    highestdamage = averagedamage
        return highestdamage / 2.0
    
    def todmodifier(self, unittype, side=0, nextturn=False):
        '''
        Return the amount unit's damage will get modified by due to
        time of day.
        
        Returns one of: 0.75, 1, 1.25.
        
        Important note! Doesn't take account of:
            * caves
            * mages of light
        '''
        if unittype.alignment == 1: # neutral
            return 1.0
        lawfulbonus = (self.gamestatus.lawful_bonus / 100.0)
        if nextturn and side <= self.myteam.side:
            # python_ai interface needs extending a bit,
            # assume standard times of day for now
            previouslawfulbonus = \
                (self.gamestatus.previous_lawful_bonus / 100.0)
            if lawfulbonus == previouslawfulbonus:
                lawfulbonus = 0.0
            elif previouslawfulbonus == 0:
                pass
            else:
                lawfulbonus = -previouslawfulbonus
        if unittype.alignment == 0: # lawful
            return 1.0 + lawfulbonus
        if unittype.alignment == 2: # chaotic
            return 1.0 - lawfulbonus
        return None # should result in an easy-to-debug exception
    
    def attackmodifiers(self, attacker, defender, attackindex, attackscore,
    fromlocation=None, tolocation=None, retaliation=False):
        '''
        Returns the approximate equivalent damage added, and damage prevented
        due to the special abilities of this attack.
        '''
        # this is full of painful approximations :'(
        attackertype = attacker.type()
        defendertype = defender.type()
        attacktype = attacker.attacks()[attackindex]
        damageadded = 0.0
        damageprevented = 0.0
        # poison
        if (attacktype.poison \
        and not defendertype.not_living \
        and not defender.poisoned):
            if defendertype.regenerate:
                damageadded += 8
            elif tolocation and self.gamemap.is_village(tolocation):
                damageadded += 8
            else:
                damageadded += min(16, 7 + defender.hitpoints)
        # slow
        if attacktype.slow and not defender.slowed:
            slowvalue = self.valueofslow(attacktype, defendertype, \
                retaliation=retaliation)
            damageprevented += slowvalue
            # also useful to prevent fast moving units from running away
            if defender.hitpoints * 3 > defender.max_hitpoints:
                damageadded += defendertype.movement # or so
        # drain
        if attacktype.drains and not defendertype.not_living:
            damageprevented += (attackscore * 0.4)
        # backstab
        if (not retaliation and fromlocation and tolocation \
        and attacktype.backstab):
            oppositetile = self.getopposite(fromlocation, tolocation)
            if oppositetile in self.allunits:
                if not self.allunits[oppositetile].is_enemy:
                    damageadded += attackscore
        return damageadded, damageprevented
    
    def attackstats(self, attacker, defender, fromtile, totile, \
    attackindex, nextturn=False, coordinated=False, lawfulbonus=0):
        '''
        Returns the following:
            * chance of killing the unit
            * chance of dying
            * expected damage given (may be negative)
            * expected damage taken (may be negative)
            * max damage given
            * max damage taken
        
        If nextturn is True, the calculations will be for the time of day
        next turn if attacker.side <= self.myteam.side.
        Otherwise if lawfulbonus is given, it's assumed that the time of
        day is such that this lawful bonus applies. Should be either 0.25
        or -0.25.
        '''
        attackertype = attacker.type()
        attackerctbh = attacker.defense_modifier(fromtile) / 100.0
        defendertype = defender.type()
        defenderctbh = defender.defense_modifier(totile) / 100.0
        attacktype = attacker.attacks()[attackindex]
        attackscore, maxattackscore = self.attackscore(attackertype, \
            defendertype, defenderctbh, attackindex, nextturn=nextturn,
            side=attacker.side)
        if lawfulbonus:
            attackscore *= 1.0 - lawfulbonus * (attackertype.alignment - 1)
        retaliationindex, retaliationscore, maxretaliationscore = \
            self.bestretaliation(attackertype, defendertype, \
            attackerctbh, attacktype, nextturn=nextturn, side=attacker.side)
        attackadded, attackprevented = self.attackmodifiers(attacker, \
            defender, attackindex, attackscore, fromtile, totile)
        if retaliationindex is not None:
            retaliationtype = defender.attacks()[retaliationindex]
            retaliationscore *= \
                1.0 - lawfulbonus * (defendertype.alignment - 1)
            retaliationadded, retaliationprevented = self.attackmodifiers( \
                defender, attacker, retaliationindex, retaliationscore, \
                totile, fromtile, retaliation=True)
        else:
            retaliationadded = retaliationprevented = 0.0
        # at the moment only prevented damage from specials is included
        # in the score for the CURRENT fight. Added damage from specials
        # may need to be included in the future, if some specials affect it,
        # but then poison will need to be factored out.
        killscore = attackscore - retaliationprevented
        deathscore = retaliationscore - attackprevented
        # this is horribly approximate :'(
        if (attacktype.berserk \
        or (retaliationindex and retaliationtype.berserk)):
            # needs it's own special calculation set
            killchance = killscore / defender.hitpoints
            deathchance = deathscore / attacker.hitpoints
            total = float(killchance + deathchance)
            killchance /= total
            deathchance /= total
            damagegiven = defender.hitpoints * killchance
            damagetaken = attacker.hitpoints * deathchance
            # ignoring poison. like a beserker would even care.
            # Slow on the other hand is accounted for already.
            maxdamagegiven = defender.hitpoints
            maxdamagetaken = attacker.hitpoints
        else:
            hitchance = self.hitchance(attacktype, defenderctbh)
            if attackscore < defender.hitpoints * hitchance:
                # (attack score previously accounted for hitchance)
                killchance = 0.0
            elif killscore <= 0:
                killchance = 0.0
            else:
                # doesn't account for strike order or deathchance
                eachattack = killscore / hitchance / attacktype.num_attacks
                attacksneeded = math.ceil(defender.hitpoints / eachattack)
                killchance = hitchance / float(attacksneeded)
            if retaliationindex is None:
                deathchance = 0.0
            else:
                hitchance = self.hitchance(retaliationtype, attackerctbh)
                if retaliationscore < attacker.hitpoints * hitchance:
                    # (retaliationscore previously accounted for hitchance)
                    deathchance = 0.0
                elif deathscore <= 0:
                    deathchance = 0.0
                else:
                    # doesn't account for strike order or killchance
                    eachattack = (deathscore / hitchance / 
                        retaliationtype.num_attacks)
                    attacksneeded = math.ceil(attacker.hitpoints / eachattack)
                    deathchance = hitchance / float(attacksneeded)
            damagegiven = attackscore + attackadded - retaliationprevented
            damagetaken = retaliationscore + retaliationadded - attackprevented
            maxdamagegiven = maxattackscore + attackadded
            maxdamagetaken = maxretaliationscore + retaliationadded
            if maxdamagegiven == damagegiven:
                # this shouldn't happen ffs
                eprint("what? max damage == damage...")
                eprint("dg, dt, maxdg, maxdt, kc, dt = " + \
                    str([damagegiven, damagetaken, maxdamagegiven, \
                    maxdamagetaken, killchance, deathchance]))
                eprint("as, aplus, aminus, ds, dplus, dminus = " + \
                    str([attackscore, attackadded, retaliationprevented, \
                    retaliationscore, retaliationadded, attackprevented]))
                eprint("hitchance, basehitchance = %s, %s" % (hitchance, \
                    defenderctbh))
                eprint("attacker %s with attacktype %s" % \
                    (self.getname(attacker), attacktype.name))
                eprint("defender %s" % self.getname(defender))
            if self.gamemap.is_village(totile) or defendertype.regenerate:
                # healers not counted atm
                if ((defender.max_hitpoints - 6) < \
                (defender.hitpoints - damagegiven)):
                    # seriously, don't even bother.
                    damagegiven = 0
                else:
                    if (not attacker.is_enemy \
                    and not totile in self.attackedlocations):
                        # throwing spears at trees is bloody useless -.-
                        # That's right, I'm talking to YOU, SPEARMAN
                        damagegiven = max(0.0, damagegiven - 2.0)
        
        return killchance, deathchance, damagegiven, damagetaken, \
            maxdamagegiven, maxdamagetaken
    
    def makepositive(self, value):
        '''
        Chances a value which may take positive and negative values
        centered on zero into a value greater than zero.
        Currently adds 1.0 for values >= 0, and returns
        1.0 / (1.0 - value) for negative values. 
        '''
        # this function doesn't really need to be smooth, only continuous
        if value >= 0.0:
            value += 1.0
        else:
            value = 1.0 / (1.0 - value)
        return value
    
    def doleader(self, mylocation, myunit, myname):
        '''
        Decide what a leader unit should do.
        
        After leader specific actions, the normal unit action function is
        called for the leader, but with the nearest keep as the main target
        in stead of the enemy leader. So it shouldn't move much unless
        the map baits it with villages and keeps.
        '''
        acted = False
        resetturn = False
        if not myunit.is_valid:
            print "Oh no! Our beloved leader %s " % myname + \
                  " has perished!?"
            return acted, resetturn
        
        if not self.contactmade:
            if self.leader_initialvillagecapture(mylocation, myunit):
                # returns True if any unit was moved. In some cases (e.g.
                # if a unit was interrupted unexpectedly) more villages can
                # be captured, so reset the turn immediately.
                return True, True
        
        self.docombatscores()
        
        self.leader_recruit(mylocation, myunit)
        
        self.leader_giveorders(mylocation, myunit)
        
        # do normal unit stuff too
        acted, resetturn = self.dounit(mylocation, myunit, myname) or acted
        
        # leader might have moved to a keep, but if so, it will recruit
        # next round.
        
        return acted, resetturn
    
    def leader_initialvillagecapture(self, mylocation, myunit):
        '''
        Work out which villages should belong to this team and send
        units towards them in the most efficient fashion possible.
        
        Returns True if any unit moved, False otherwise.
        '''
        myname = self.getname(myunit)
        friendlybydest = wesnoth.get_units_by_destination()
        friendlybysource = wesnoth.get_destinations_by_unit()
        somethingmoved = None
        if not self.contactmade:
            # check if we're in range of the enemy yet
            enemiesbydest = wesnoth.get_enemy_units_by_destination()
            for destinations in friendlybysource.itervalues():
                for destination in destinations:
                    if destination in enemiesbydest:
                        self.contactmade = True
                        return False
        dprint("Looking for villages to capture")
        leaders = self.getleaders()
        snatchlater = []
        snatchnow = []
        for village in self.villages:
            villageowned = None
            dibs = False
            mydistance = mylocation.distance_to(village)
            for team in wesnoth.get_teams():
                if team.owns_village(village):
                    villageowned = True
                    break
                for location, leader in leaders:
                    if location == mylocation:
                        continue
                    if location.distance_to(village) < mydistance:
                        dibs = True
                        break
                if dibs: # someone else is closer to it
                    villageowned = True
                    break
            if not villageowned: # we WANTS IT
                if village in friendlybydest:
                    snatchnow.append(village)
                else:
                    snatchlater.append(village)
        
        def snatch(village, snatcher):
            '''
            Capture a village.
            '''
            if snatcher not in self.allunits:
                eprint("..>..>...??? no unit at %s D:D:D:D:" % \
                    printable(snatcher))
            dprint("%s orders %s to capture a village at %s" % (myname,
                self.getname(self.allunits[snatcher]), printable(village)))
            doublecheck = wesnoth.move_unit(snatcher, village)
            if not doublecheck == village:
                # maybe there was a unit hiding on it?
                # maybe someone implemented a new-unit-visible interrupt!
                # maybe the sun shines out of my ass
                wprint("something unexpected happened while trying" + \
                    "to capture a village at %s!" % printable(village))
                wprint("It seems the whole turn should be re-thought out!")
                return False
            return True
        
        def atmovelimit(sometile, unitlocation, reachabletiles):
            '''
            Check if a unit can't move further than some tile.
            '''
            basedistance = unitlocation.distance_to(sometile)
            surroundings = wesnoth.get_adjacent_tiles(sometile)
            for tile in surroundings:
                if (tile.distance_to(unitlocation) > basedistance \
                and tile in reachabletiles):
                    return False
            return True
        
        # first see how many villages we can snatch immediately.
        # Try to capture them with the unit that uses the most movement
        # to get there.
        takencareof = []
        for village in snatchnow:
            if village not in friendlybydest:
                # the snatcher snatched something else already
                snatchlater.append(village)
                continue
            potentialsnatchers = friendlybydest[village]
            if mylocation in potentialsnatchers:
                # you probably have better things to be doing
                potentialsnatchers.remove(mylocation)
            if len(potentialsnatchers) == 1:
                # SNATCH IT
                snatcher = potentialsnatchers[0]
                okay = snatch(village, snatcher)
                if not okay:
                    return not okay # tee hee
                takencareof.append(village)
                friendlybydest = wesnoth.get_units_by_destination()
                friendlybysource = wesnoth.get_destinations_by_unit()
                somethingmoved = True
                continue
            gotit = False
            for location in potentialsnatchers:
                snatcher = self.allunits[location]
                reachabletiles = friendlybysource[location]
                if atmovelimit(village, location, reachabletiles):
                    # SNATCH IT
                    okay = snatch(village, location)
                    if not okay:
                        return not okay
                    gotit = True
                    break
            if gotit:
                friendlybydest = wesnoth.get_units_by_destination()
                friendlybysource = wesnoth.get_destinations_by_unit()
                somethingmoved = True
                continue
            for location in potentialsnatchers:
                # how many other villages can this unit reach?
                villagecount = 0
                for tile in reachabletiles:
                    if (self.gamemap.is_village(tile) \
                    and tile not in takencareof):
                        villagecount += 1
                if villagecount == 1:
                    # can only reach this village, so...
                    # yeah you guessed it :-D
                    okay = snatch(village, location)
                    if not okay:
                        return not okay
                    gotit = True
                    break
            if gotit:
                friendlybydest = wesnoth.get_units_by_destination()
                friendlybysource = wesnoth.get_destinations_by_unit()
                somethingmoved = True
                continue
            # TODO: actually this is all for now.
            # I would really like to take account of all units, and set
            # goto orders for villages more than one move away,
            # but it is a lot of work, it's late, and I'm tired.
            # Some other time!
        
        return somethingmoved
    
    def docombatscores(self):
        '''
        Calculate combat scores for each recruitable type, weighed
        against all enemy units.
        
        Keys are unit type names, as unittype objects are not hashable.
        
        Hidden units, or units in fog / shroud are NOT included.
        '''
        wesnoth.raise_user_interact()
        allunits = wesnoth.get_units()
        myunits = wesnoth.get_destinations_by_unit()
        visibleenemies = wesnoth.get_enemy_destinations_by_unit()
        self.combatscores = {}
        myunittypes = [allunits[location].type() for location in myunits]
        for unittype in (myunittypes + self.myrecruittypes):
            wesnoth.raise_user_interact()
            if unittype.name in self.combatscores:
                continue
            bestlawfulbonus = -0.25 * (unittype.alignment - 1)
            self.combatscores[unittype.name] = 1.0
            enemytotals = {}
            for enemylocation in visibleenemies:
                enemy = allunits[enemylocation]
                enemytype = enemy.type()
                if enemytype.name in enemytotals:
                    continue
                # work out offensive and defensive scores vs. this enemy
                bestenemylawfulbonus = (-0.25 * (enemytype.alignment - 1)) \
                    or -bestlawfulbonus
                if not bestlawfulbonus:
                    # if attacker is neutral, choose the lawful bonus
                    # that's worst for the enemy
                    bestlawfulbonus = -bestenemylawfulbonus
                if unittype.name in self.defensescores:
                    myctbh = 1.0 - self.defensescores[unittype.name]
                else:
                    # one of the units leveled up during this turn
                    myctbh = 0.5
                if enemytype.name in self.defensescores:
                    enemyctbh = 1.0 - self.defensescores[enemytype.name]
                else:
                    # an enemy leveled up during this turn
                    # i.e. because of our attack.
                    enemyctbh = 0.5
                
                offensivescore, bestattacktype = self.combatscore(unittype, \
                    enemytype, myctbh, enemyctbh, lawfulbonus=bestlawfulbonus)
                
                enemyoffensivescore, bestenemyattacktype = self.combatscore( \
                    enemytype, unittype, enemyctbh, myctbh, \
                    lawfulbonus=bestenemylawfulbonus)
                
                defensivescore = -enemyoffensivescore
                bestattackname = None
                bestenemyattackname = None
                if bestattacktype:
                    bestattackname = bestattacktype.name
                if bestenemyattacktype:
                    bestenemyattackname = bestenemyattacktype.name
                megadprint("%s vs. %s: best offense = %s, score %s" % \
                    (unittype.name, enemytype.name, bestattackname, \
                    offensivescore))
                megadprint("%s vs. %s: best enemy offense = %s, score %s" % \
                    (unittype.name, enemytype.name, bestenemyattackname, \
                    -defensivescore))
                enemytotals[enemytype.name] = [offensivescore, defensivescore]
            self.combatscores[unittype.name] = enemytotals
        return True
    
    def combatscore(self, attackertype, defendertype, attackerctbh, \
    defenderctbh, lawfulbonus=0):
        '''
        Find the most effective attack vs. defendertype with defenderctbh
        and return a measure of how effective this attack is, along with
        the attack type.
        
        Basic formula is:
            effectiveness = damagegiven * unit cost / unit hp
            effectiveness -= damagetaken * my cost / my hp
        
        if 'lawfulbonus' is given, it will be assumed that the time of day
        is such that this lawful bonus is applied.
        '''
        # NOTE: attackscore should be inversely proportional
        # to the number of turns required to kill the enemy.
        # It will also be proportional to the cost damage
        # inflicted, and inversely proportional to the
        # cost damage received in retaliation.
        attacks = attackertype.attacks()
        effectiveness = None
        bestattacktype = None
        for attacktype in attacks:
            attackindex = attacks.index(attacktype)
            attackscore, unused, attackadded, attackprevented = \
                self.attackscore( \
                attackertype, defendertype, defenderctbh, attackindex, \
                generic=True, lawfulbonus=lawfulbonus)
            retaliationindex, retaliationscore, unused, \
            retaliationadded, retaliationprevented = \
                self.bestretaliation(attackertype, defendertype, attackerctbh,
                attacktype, generic=True, lawfulbonus=lawfulbonus)
            if attacktype.charge:
                attackscore *= 2.5 # (attacker gets first hit)
                retaliationscore *= 2.0
            if retaliationindex is None:
                retaliationtype = None
            else:
                retaliationtype = defendertype.attacks()[retaliationindex]
            if attacktype.berserk and retaliationtype is None:
                # DARK ADEPT SLAUGHTERFEST!!!
                score = defendertype.cost
            elif (attacktype.berserk \
            or (retaliationtype and retaliationtype.berserk)):
                attackscore -= retaliationprevented
                retaliationscore -= attackprevented
                if attackscore < 0:
                    attackscore = 0
                if retaliationscore < 0:
                    retaliationscore = 0
                attackscore /= float(defendertype.hitpoints)
                killscore = attackscore
                killscore *= defendertype.cost
                retaliationscore /= float(attackertype.hitpoints)
                deathscore = retaliationscore
                deathscore *= attackertype.cost
                if attackscore == retaliationscore:
                    # this is only really for the extremely rare 0, 0 case
                    score = defendertype.cost - attackertype.cost
                else:
                    score = (killscore - deathscore) / \
                        (attackscore + retaliationscore)
            else:
                attackscore += attackadded - retaliationprevented
                attackscore /= float(defendertype.hitpoints)
                attackscore *= defendertype.cost
                retaliationscore += retaliationadded - attackprevented
                retaliationscore /= float(attackertype.hitpoints)
                retaliationscore *= attackertype.cost
                score = attackscore - retaliationscore
            if (effectiveness is None \
            or score > effectiveness):
                effectiveness = score
                bestattacktype = attacktype
        
        return effectiveness, bestattacktype
    
    def leader_recruit(self, mylocation, myunit):
        '''
        Choose which units to recruit, and recruit them.
        '''
        allunits = wesnoth.get_units()
        leaders = self.getleaders()
        mytype = myunit.type()
        myname = self.getname(myunit)
        
        if not self.gamemap.is_keep(mylocation):
            return False
        
        if not self.keeps[mylocation]: # no attached castle tiles
            return False
        
        myunitsworth = {} # name -> count
        totalmyworth = 0
        enemyunitsworth = {} # name -> count
        totalenemyworth = 0
        for location, unit in allunits.iteritems():
            # note: includes all hidden and flavour units
            # e.g. units in shroud / fog, elvish avengers in forest, cows.
            # this is because I don't want the AI to feel like it's
            # already won in matches involving fog / shroud.
            # if there is a way to ignore flavour units and stealthed units,
            # this should be implemented!
            if unit.stoned: # ignore statues
                continue
            unittype = unit.type()
            unitworth = self.getworth(unit)
            basecost = unittype.cost
            if unit.is_enemy:
                totalenemyworth += unitworth
                if not unittype.name in enemyunitsworth:
                    enemyunitsworth[unittype.name] = unitworth
                else:
                    enemyunitsworth[unittype.name] += unitworth
                continue
            elif unit.side == self.myteam.side:
                totalmyworth += unitworth
                if not unittype.name in myunitsworth:
                    myunitsworth[unittype.name] = unitworth
                else:
                    myunitsworth[unittype.name] += unitworth
                continue
            else:
                # ignore allied units for now.
                pass
        
        self.winning = False
        ratio = totalmyworth / totalenemyworth
        if totalmyworth < 120 and totalenemyworth < 120: noun = 'team'
        elif ratio > 4:
            noun = 'unstoppable killing force'
            self.winning = True
        elif ratio > 2:
            noun = 'vast batallion'
            self.winning = True
        elif ratio > 1: noun = 'grand army'
        elif ratio <= 1: noun = 'brave regiment'
        elif ratio < 2: noun = 'puny band of rag-tag soldiers and ruffians'
        elif ratio < 4: noun = 'pathetic remaining followers'
        iprint("%s surveys its %s" % (myname, noun))
        if self.winning:
            iprint("%s shouts: " % myname + \
                '"Onward troops! Fear not the enemies\' reprisal!"')
        myvillagecount = 0
        for village in self.villages:
            if self.myteam.owns_village(village):
                myvillagecount += 1
        if myvillagecount * 2 > len(self.villages) and ratio > 4:
            # recruiting more looks like overkill. Save gold for a rainy day!
            dprint("%s likes what it sees!" % myname)
            return
        def minihelper(dictionary, multiplier, name, score):
            '''
            Inserts a score into the given dictionary indexed by unit name.
            'multiplier' is the total value of the enemy units that
            the score applies to. (so for e.g. scores vs. ghosts are more
            important if the enemy fields an army full of ghosts)
            '''
            if name:
                score *= multiplier
                if name not in dictionary:
                    dictionary[name] = score
                else:
                    dictionary[name] += score
                return score
            return 0.0
        offensiveunits = {} # name -> weight
        defensiveunits = {}
        hinderanceunits = {}
        cheapunits = {}
        totaloffensive = 0.0
        totaldefensive = 0.0
        totalhinderance = 0.0
        totalvalueformoney = 0.0
        for enemyname in enemyunitsworth:
            if enemyname not in self.movementscores:
                # can happen after level up
                continue
            enemymovementscore = self.movementscores[enemyname]
            bestoffensive = [None, 0] # name, score
            secondbestoffensive = [None, 0] # etc
            bestdefensive = [None, 0]
            secondbestdefensive = [None, 0]
            bestmovement = [None, 0]
            secondbestmovement = [None, 0]
            bestvalue = [None, 0]
            secondbestvalue = [None, 0]
            for recruittype in self.myrecruittypes:
                # note: this has reached the point where it needs a
                # REALLY BIG CLEANUP
                recruitname = recruittype.name
                offensivescore = \
                    self.combatscores[recruitname][enemyname][0]
                defensivescore = \
                    self.combatscores[recruitname][enemyname][1]
                # for defensivescores the values are almost always negative
                # so it has to be accounted for... assume that net damage
                # (damage taken - damage given) of less than 1/4 of our
                # cost is above average (yes this is completely made up)
                defensivescore += (recruittype.cost / 4.0)
                if enemymovementscore:
                    movementscore = (self.movementscores[recruitname] - \
                        float(enemymovementscore))
                    # note: don't normalise by cost at all. cheaper units
                    # are more likely to die, thus on average, hinderance
                    # is probably decreased by a smiliar amount to the
                    # cost savings.
                else:
                    movementscore = 0
                if not recruittype.has_zoc:
                    # this ensures that for e.g. ghosts are counted
                    # as having better hinderance than bats vs. 
                    # low-movement units.
                    # TODO: test for enemy skirmishers... can't atm.
                    movementscore /= 2.0
                cost = float(recruittype.cost)
                valueformoney = ((offensivescore + defensivescore + \
                    movementscore) / 3.0)
                if offensivescore > bestoffensive[1]:
                    secondbestoffensive = bestoffensive
                    bestoffensive = [recruitname, offensivescore]
                elif offensivescore > secondbestoffensive[1]:
                    secondbestoffensive = [recruitname, offensivescore]
                if defensivescore > bestdefensive[1]:
                    secondbestdefensive = bestdefensive
                    bestdefensive = [recruitname, defensivescore]
                elif defensivescore > secondbestdefensive[1]:
                    secondbestdefensive = [recruitname, defensivescore]
                if movementscore > bestmovement[1]:
                    secondbestmovement = bestmovement
                    bestmovement = [recruitname, movementscore]
                elif movementscore > secondbestmovement[1]:
                    secondbestmovement = [recruitname, movementscore]
                if valueformoney > bestvalue[1]:
                    secondbestvalue = bestvalue
                    bestvalue = [recruitname, valueformoney]
                elif valueformoney > secondbestvalue[1]:
                    secondbestvalue = [recruitname, valueformoney]
            megadprint("vs. %s: best offensive: %s with score %g" % \
                (enemyname, bestoffensive[0], bestoffensive[1]))
            megadprint("vs. %s: second best: %s with score %g" % \
                (enemyname, secondbestoffensive[0], secondbestoffensive[1]))
            megadprint("vs. %s: best defensive: %s with score %g" % \
                (enemyname, bestdefensive[0], bestdefensive[1]))
            megadprint("vs. %s: scond best: %s with score %g" % \
                (enemyname, secondbestdefensive[0], secondbestdefensive[1]))
            megadprint("vs. %s: best hinderance: %s with score %g" % \
                (enemyname, bestmovement[0], bestmovement[1]))
            megadprint("vs. %s: second best: %s with score %g" % \
                (enemyname, secondbestmovement[0], secondbestmovement[1]))
            megadprint("vs. %s: best value for money: %s with score %g" % \
                (enemyname, bestvalue[0], bestvalue[1]))
            megadprint("vs. %s: second best: %s with score %g" % \
                (enemyname, secondbestvalue[0], secondbestvalue[1]))
            enemyworth = enemyunitsworth[enemyname]
            totaloffensive += \
                minihelper(offensiveunits, enemyworth, *bestoffensive)
            totaloffensive += \
                minihelper(offensiveunits, enemyworth, *secondbestoffensive)
            totaldefensive += \
                minihelper(defensiveunits, enemyworth, *bestdefensive)
            totaldefensive += \
                minihelper(defensiveunits, enemyworth, *secondbestdefensive)
            totalhinderance += \
                minihelper(hinderanceunits, enemyworth, *bestmovement)
            totalhinderance += \
                minihelper(hinderanceunits, enemyworth, *secondbestmovement)
            totalvalueformoney += \
                minihelper(cheapunits, enemyworth, *bestvalue)
            totalvalueformoney += \
                minihelper(cheapunits, enemyworth, *secondbestvalue)
        finalweights = {}
        fuzz = (len(enemyunitsworth) < 3) # if < 3 enemies, fuzz our weights
        healers = {}
        totalhealing = 0.0
        for recruittype in self.myrecruittypes:
            recruitname = recruittype.name
            healing = 0.0
            if recruittype.curing:
                healing = 10.0 / float(recruittype.cost)
            elif recruittype.heals:
                healing = 4.0 / float(recruittype.cost)
            if healing:
                healers[recruitname] = healing
                totalhealing += healing
        totalfinalweight = 0.0
        reasons = {}
        for recruittype in self.myrecruittypes:
            recruitname = recruittype.name
            finalweights[recruitname] = 0.0
            reasons[recruitname] = []
            if recruitname in offensiveunits:
                offensiveweight = \
                    offensiveunits[recruitname] / float(totaloffensive)
                finalweights[recruitname] += offensiveweight
                reasons[recruitname].append(('o', offensiveweight))
            if recruitname in defensiveunits:
                defensiveweight = \
                    defensiveunits[recruitname] / float(totaldefensive)
                finalweights[recruitname] += defensiveweight
                reasons[recruitname].append(('d', defensiveweight))
            if recruitname in hinderanceunits:
                hinderanceweight = \
                    hinderanceunits[recruitname] / float(totalhinderance)
                finalweights[recruitname] += hinderanceweight
                reasons[recruitname].append(('h', hinderanceweight))
            if recruitname in cheapunits:
                valueweight = \
                    cheapunits[recruitname] / float(totalvalueformoney)
                finalweights[recruitname] += valueweight
                reasons[recruitname].append(('v', valueweight))
            consideredfactors = 4.0
            if totalhealing:
                consideredfactors += 1.0
                if recruitname in healers:
                    finalweights[recruitname] += \
                        healers[recruitname] / float(totalhealing)
                    reasons[recruitname].append('heals')
            if fuzz:
                consideredfactors += 2.0
                finalweights[recruitname] += \
                    2.0 / float(len(self.myrecruittypes))
            finalweights[recruitname] /= float(consideredfactors)
            dprint("final weight for %s: %g" % \
                (recruitname, finalweights[recruitname]))
            dprint("reasons: " + str(reasons[recruitname]))
            totalfinalweight += finalweights[recruitname]
        # here's to hoping that i didn't fuck up.
        # these motherfuckers should be perfectly weighted.
        megadprint("total weights add to %s" % totalfinalweight)
        if abs(totalfinalweight - 1.0) > 0.000001:
            eprint("!!! ERROR !!! recruitment weights should add to 1.0!")
        
        recruitmore = True
        while recruitmore:
            wesnoth.raise_user_interact()
            # this will just keep recruiting until it either can't buy its
            # weighted random choice, or runs out of castle tiles.
            recruitmore = False
            currentunits = {}
            totalunits = 0
            for location, unit, name in self.getmyunits():
                typename = unit.type().name
                if typename not in finalweights:
                    # only consider units we can actually recruit
                    # ignores lvl 2's, hero units, etc.
                    # i think this is fine
                    continue
                if typename not in currentunits:
                    currentunits[typename] = 1
                else:
                    currentunits[typename] += 1
                totalunits += 1
            totalunits = float(totalunits)
            bestchoice = None
            bestscore = 0.0
            for name in finalweights:
                if name in currentunits:
                    currentweight = currentunits[name] / totalunits # float
                else:
                    currentweight = 0.0
                desiredweight = finalweights[name]
                if not desiredweight:
                    continue
                if currentweight <= desiredweight:
                    score = desiredweight - currentweight
                    if score >= bestscore:
                        # if all else is equal, will recruit the last unit
                        # it comes across... e.g. if all weights are 0 etc.
                        bestscore = score
                        bestchoice = name
            if not bestchoice:
                wprint("%s couldn't find anything useful to recruit" % myname)
                break
            o, d, h = 0, 0, 0
            if bestchoice in offensiveunits:
                o = offensiveunits[bestchoice] / float(totaloffensive)
            if bestchoice in defensiveunits:
                d = defensiveunits[bestchoice] / float(totaldefensive)
            if bestchoice in hinderanceunits:
                h = hinderanceunits[bestchoice] / float(totalhinderance)
            for castletile in self.keeps[mylocation]:
                # on maps with a lot of castle tiles this can get slow :(
                #wesnoth.raise_user_interact()
                if wesnoth.recruit_unit(bestchoice, castletile):
                    dprint("%s recruited a %s to fix an imbalance of %g%%" % \
                        (myname, bestchoice, bestscore * 100))
                    megadprint("%s has offensive, defensive, hinderance" % \
                        bestchoice + " weights of: %g, %g, %g" % (o, d, h))
                    self.recruitedthisturn = True
                    recruitmore = True
                    break
    
    def leader_giveorders(self, mylocation, myunit):
        '''
        Look for objectives requiring teamwork, and instruct units to
        attempt to complete them.
        
        Units can choose to ignore the orders, if they think they're stupid.
        '''
        myunitsbydest = wesnoth.get_units_by_destination()
        enemiesbysource = wesnoth.get_enemy_destinations_by_unit()
        leadername = self.getname(myunit)
        dprint("%s squints intently at the battlefield." % leadername)
        enemyheldlocations = {}
        for location in self.criticallocations:
            if (location in enemiesbysource \
            and self.allunits[location].is_enemy):
                enemyheldlocations[location] = self.criticallocations[location]
        vulnerablelist = []
        myusefulunits = []
        possibleattackers = {}
        mybackstabbers = []
        
        for enemylocation in enemiesbysource:
            # NOTE: enemiesbysource doesn't include hidden / shrouded units
            # allenemies does
            enemy = self.allenemies[enemylocation]
            enemytype = enemy.type()
            enemyctbh = enemy.defense_modifier(enemylocation) / 100.0
            # which units can reach it?
            # this is a little approximate, but fairly efficient
            # just takes the unit which can do the most damage from each tile
            # and adds their max damage (hits 100%, enemy hits 0%)
            possibleattackers[enemylocation] = {}
            attackers = []
            totalmaxdamage = 0.0
            for tile in wesnoth.get_adjacent_tiles(enemylocation):
                if tile in myunitsbydest:
                    possibleattackers[enemylocation][tile] = []
                    tilemaxdamage = 0.0
                    bestattackunit = None
                    for unitlocation in myunitsbydest[tile]:
                        if unitlocation == mylocation:
                            # don't give yourself orders
                            continue
                        if self.gamemap.is_village(unitlocation):
                            # don't order units to leave important things
                            # undefended if you're too stupid to order
                            # them to defend such important things
                            continue
                        unit = self.allunits[unitlocation]
                        if unit.type().usage == 'scout':
                            # scouts are better left to their own devices
                            continue
                        ctbh = unit.defense_modifier(tile) / 100.0
                        if ctbh > 0.65:
                            # for now: just don't.
                            continue
                        indexfound = False
                        for index in xrange(len(myusefulunits)):
                            if myusefulunits[index][1] == unitlocation:
                                myusefulunits[index][0] += 1
                                myusefulunits[index].append(\
                                    (tile, enemylocation))
                                indexfound = True
                                break
                        if not indexfound:
                            index = len(myusefulunits)
                            myusefulunits.append(\
                                [1, unitlocation, (tile, enemylocation)])
                        possibleattackers[enemylocation][tile].append( \
                            unitlocation)
                        if not unitlocation in attackers:
                            maxdamage, attackindex = self.maxdamage( \
                                unit.type(), enemytype)
                            if maxdamage > tilemaxdamage:
                                tilemaxdamage = maxdamage
                                bestattackunit = unitlocation
                    if bestattackunit:
                        totalmaxdamage += tilemaxdamage
                        attackers.append(bestattackunit)
            if totalmaxdamage >= enemy.hitpoints:
                opportunity = totalmaxdamage / float(enemy.hitpoints)
                # assume critical locations are worth 8g (needs tweaking)
                if enemylocation in self.criticallocations:
                    tilebonus = 8
                else:
                    tilebonus = 0
                opportunity *= (self.getworth(enemy) + tilebonus)
                opportunity *= enemyctbh
                vulnerablelist.append([opportunity, enemylocation, enemy])
        # units themselves will watch out for other wounded units
        # when it is their turn to move. So protecting them via zoc here
        # isn't necessary.
        # plus they can run away on their own. cowards!
        # What should be considered is saving wounded comrades by
        # killing enemies threatening them... but i'm not sure whether
        # this should be a leader's role, or part of the unit's
        # individual decision.
        
        # TODO: Leader should consider critical locations which might
        # be lost next turn, and either protect them, yield them, or
        # leave it to the unit's discretion.
        
        # TODO: look at special units, such as berserkers and figure out
        # how to effectively counter them (e.g. in this case, either kill
        # them before they kill you, or weaken them with cannon fodder)
        
        # considering which unit should do what is going to be extremely
        # complicated. I'm using my own decision-making process as a vague
        # basis for this, starting off really simple.
        
        # i'm gonna do this in this order:
        # 1) decide what units should do who can only do 1 useful thing
        # 2) gang up on enemy units with the remaining ones, starting with
        # the most valuable killable enemy
        
        # atm the leader is pretty stupid. doesn't even consider if his
        # units are hurt. he'll only get cleverer. Look out for yourselves
        # little units!
        
        self.orders = {}
        orderedunits = []
        usedlocations = []
        
        vulnerablelist.sort(reverse=True)
        myusefulunits.sort()
        for opportunity, enemylocation, enemy in vulnerablelist:
            enemytype = enemy.type()
            if len(orderedunits) >= len(myusefulunits):
                # looks like we're done
                break
            # technique here:
            # * first attack with those that can't attack anything else
            # * soften it up with units that take little damage attacking
            # * finish it off with units that have higher xp
            # * don't waste xp
            potentialxp = self.xpfromunit(enemy)
            top2byxp = [[0, None], [0, None]]
            allbycombatscore = []
            expecteddamage = {}
            for actionlist in myusefulunits:
                unitlocation = actionlist[1]
                if unitlocation in orderedunits:
                    continue
                unit = self.allunits[unitlocation]
                unittype = unit.type()
                possiblexpgain = unit.max_experience - unit.experience
                wastedxp = potentialxp - possiblexpgain
                unithp = unit.hitpoints / float(unit.max_hitpoints)
                wounded = 0.5 - unithp
                if wastedxp > 3:
                    if wounded:
                        xpscore = 20
                    else:
                        xpscore = 4 - wastedxp
                elif wastedxp > -1:
                    xpscore = 20 - abs(wastedxp)
                elif wastedxp > -15:
                    xpscore = 15 + wastedxp
                else:
                    xpscore = 0
                if xpscore > top2byxp[0][0]:
                    top2byxp[1] = top2byxp[0]
                    top2byxp[0] = [xpscore, unitlocation]
                elif xpscore > top2byxp[1][0]:
                    top2byxp[1] = [xpscore, unitlocation]
                if not enemytype.name in self.combatscores[unittype.name]:
                    # this shouldn't really ever happen
                    dprint("she sells sea shells by the she shore... !@#$")
                offensivescore, defensivescore = \
                    self.combatscores[unittype.name][enemytype.name]
                if offensivescore > 0:
                    allbycombatscore.append([offensivescore, unitlocation])
            allbycombatscore.sort(reverse=True)
            def dobyxp(damagedone):
                expectedenemyhp = enemy.hitpoints - damagedone
                for byxp in top2byxp:
                    unitlocation = byxp[1]
                    if not unitlocation or (unitlocation in orderedunits):
                        continue
                    unit = self.allunits[unitlocation]
                    unittype = unit.type()
                    attackindex, attackscore, retaliationscore = \
                       self.safestattack(unittype, enemytype)
                    if attackscore * enemyctbh > enemy.hitpoints:
                        # kill it!
                        foundtile = None
                        bestctbh = 1.0
                        reasons = []
                        for tile in possibleattackers[enemylocation]:
                            if (unitlocation not in \
                            possibleattackers[enemylocation][tile]):
                                continue
                            if tile in usedlocations:
                                continue
                            attackstats = self.attackstats(unit, enemy, tile, \
                                enemylocation, attackindex, coordinated=True)
                            deathchance = attackstats[1]
                            if deathchance:
                                reasons.append('it could perish')
                                # change your mind!
                                continue
                            damagegiven = attackstats[2]
                            expectedenemyhp = enemy.hitpoints - damagedone
                            if not damagegiven > expectedenemyhp:
                                reasons.append("it probably wouldn't kill it")
                                # maybe someone else will hurt it more first
                                continue
                            if not foundtile:
                                foundtile = tile
                            else:
                                ctbh = unit.defense_modifier(tile) / 100.0
                                if ctbh < bestctbh:
                                    foundtile = tile
                                    bestctbh = ctbh
                        if foundtile:
                            dprint("%s orders %s to kill %s" % (leadername, \
                                self.getname(unit), self.getname(enemy)))
                            self.orders[unitlocation] = ['attack', \
                                enemylocation, 'mustkill']
                            orderedunits.append(unitlocation)
                            usedlocations.append(foundtile)
                            return True
                        else:
                            if not reasons:
                                reasons = ["%s can't find " % leadername + \
                                    "anywhere to attack from"]
                            dprint("high xp unit %s " % self.getname(unit) + \
                                "won't attack %s, because " % self.getname(enemy) + \
                                ' and '.join(reasons))
            if dobyxp(0):
                # our high xp guy's got it yo
                continue
            totaldamagedone = 0.0
            for score, unitlocation in allbycombatscore:
                if unitlocation in orderedunits:
                    continue
                unit = self.allunits[unitlocation]
                unittype = unit.type()
                attackindex, attackscore, retaliationscore = \
                    self.safestattack(unittype, enemytype)
                bestdamagetaken = None
                foundtile = None
                reasons = []
                for tile in possibleattackers[enemylocation]:
                    if (not unitlocation in \
                    possibleattackers[enemylocation][tile]):
                        continue
                    if tile in usedlocations:
                        continue
                    if tile in self.allunits and tile != unitlocation:
                        blockerunit = self.allunits[tile]
                        if blockerunit in orderedunits:
                            if self.orders[tile][1] == tile:
                                reasons.append(printable(tile) + \
                                    " someone else is busy there")
                                continue
                            else: # we're ok, it's gonna move
                                pass
                        else:
                            reasons.append("%s there's someone in the way" % \
                                printable(tile))
                            continue
                    attackstats = self.attackstats(unit, enemy, tile, \
                        enemylocation, attackindex, coordinated=True)
                    deathchance = attackstats[1]
                    if deathchance > 0.05:
                        reasons.append("%s it might die" % printable(tile))
                        # change your mind
                        continue
                    damagegiven = attackstats[2]
                    damagetaken = attackstats[3]
                    if (bestdamagetaken is None \
                    or (damagetaken < bestdamagetaken)):
                        bestdamagetaken = damagetaken
                        foundtile = tile
                        damagedone = damagegiven
                if foundtile:
                    dprint("%s orders %s to assault %s" % (leadername, \
                        self.getname(unit), self.getname(enemy)))
                    self.orders[unitlocation] = \
                        ['attack', enemylocation]
                    orderedunits.append(unitlocation)
                    usedlocations.append(foundtile)
                    totaldamagedone += damagedone
                    if dobyxp(totaldamagedone):
                        # end of the line
                        break
                else:
                    if not reasons:
                        reasons = ["it can't find anywhere to attack from"]
                    dprint("%s couldn't find a good" % leadername + \
                        " way for %s to assault %s, because " % \
                        (self.getname(unit), self.getname(enemy)) + \
                        " and ".join(reasons))
        if self.orders:
            # Abuses the fact that the leader is always sorted first in
            # self.myunits.
            newunitlist = [(mylocation, myunit, leadername)]
            for unitlocation in orderedunits:
                unit = self.allunits[unitlocation]
                name = self.getname(unit)
                newunitlist.append((unitlocation, unit, name))
            for location, unit, name in self.myunits:
                if (location not in orderedunits \
                and location != mylocation):
                    newunitlist.append((location, unit, name))
            self.myunits = newunitlist
        else:
            text = random.choice([ \
                " tells a bad joke",
                " can't think of anything motivational to say",
                " picks its nose and examines it",
                " coughs and shifts its weight uncomfortably",
                " ponders the meaning of life",
                " gives a stirring speech, " + \
                    "but no one is listening",
                ' suddenly gives a rousing battle cry! "YEAARGH!!!"',
                " looks mildly amused for some reason",
                " calls for pizza",
                " invites everyone to the pub for a drink"])
            dprint(str(leadername) + text)
            
    
    def maxdamage(self, attackertype, defendertype, side=0, \
    nextturn=False):
        '''
        Simple calculation for maximum damage one unit can do to another
        '''
        attacktypes = attackertype.attacks()
        bestscore = 0.0
        bestattackindex = None
        for attackindex in xrange(len(attacktypes)):
            score, unused = self.attackscore(attackertype, defendertype, \
                1.0, attackindex, nextturn, side)
            if score >= bestscore:
                # (some flavour units can have 0 damage attacks)
                bestscore = score
                bestattackindex = attackindex
        if bestattackindex is None:
            wprint("WARNING: can't find best attack for %s against %s!" % \
                (attackertype.name, defendertype.name))
        return bestscore, bestattackindex
    
    def safestattack(self, attackertype, defendertype):
        '''
        Return the index of the safest attack vs. a certain type, the
        max damage given and max damage taken, assuming 50% defense
        '''
        attacktypes = attackertype.attacks()
        minscore = None
        bestattackindex = None
        bestattackscore = 0.0
        for attackindex in xrange(len(attacktypes)):
            attacktype = attacktypes[attackindex]
            attackscore, unused, attackadded, attackprevented = \
                self.attackscore(attackertype, defendertype, 0.5, \
                attackindex, generic=True)
            retaliationindex, retaliationscore, maxretaliationscore, \
            retaliationadded, retaliationprevented = \
                self.bestretaliation(attackertype, defendertype, 0.5,
                attacktype, generic=True)
            retaliationscore += retaliationadded - attackprevented
            attackscore += attackadded - retaliationprevented
            if minscore is None or retaliationscore < minscore:
                minscore = retaliationscore
                bestattackindex = attackindex
                bestattackscore = attackscore
            elif retaliationscore == minscore:
                if attackscore > bestattackscore:
                    minscore = retaliationscore
                    bestattackindex = attackindex
                    bestattackscore = attackscore
        bestattackscore *= self.todmodifier(attackertype)
        retaliationscore = minscore * self.todmodifier(defendertype)
        return bestattackindex, bestattackscore, retaliationscore
        
        
    def dounit(self, mylocation, myunit, myname):
        '''
        Decide what an individual unit should do.
        
        The decision making is split into two somewhat codependant parts.
        Each unit is given a set of behaviour modifiers depending on its
        current state, and each tile that the unit can move to is scored
        against a variety of criteria.
        
        The best move to make is currently decided by weighting the scores
        differently according to the current behaviour modifiers.
        
        Effectively each unit is relying purely on instinct.
        '''
        acted = False
        resetturn = False
        doover = False
        if not myunit.is_valid:
            print "Oh, my beloved %s, " % myname + \
                  "where have you gone?"
            return acted, resetturn
        mytype = myunit.type()
        friendlymoves = wesnoth.get_destinations_by_unit()
        if mylocation in friendlymoves:
            mymoves = friendlymoves[mylocation]
            if not mylocation in mymoves:
                # the default AI doesn't consider staying still unless it
                # has full movement points.
                mymoves.append(mylocation)
                # it also removes allied villages from the move list
                # but for now this is okay
        else:
            mymoves = []
            if not myunit.can_attack:
                return acted, resetturn
        friendlyunits = wesnoth.get_units_by_destination()
        if not mymoves:
            adjacentenemies = False
            for tile in wesnoth.get_adjacent_tiles(mylocation):
                if (tile in self.allunits.keys() \
                and self.allunits[tile].is_enemy):
                    adjacentenemies = True
                    break
            if not adjacentenemies:
                return acted, resetturn
        if mylocation in self.orders:
            myorders = self.orders[mylocation]
        else:
            myorders = None
        
        feelings = self.unit_feelings(mylocation, myunit)
        
        primarytargets, secondarytargets = \
            self.unit_targets(mylocation, myunit)
        
        movestats = self.unit_movestats(mylocation, myunit, mymoves, \
            primarytargets, secondarytargets)
        
        deathvalue = self.getworth(myunit) + \
            self.xptogold(self.xpfromunit(myunit))
        despair = 0.95 * deathvalue
        ultimatedespair = 0.99 * deathvalue
        for move in movestats:
            if move['deathchance'] < despair:
                despair = 0.0
                ultimatedespair = 0.0
                break
            if ultimatedespair and move['deathchance'] < ultimatedespair:
                ultimatedespair = 0.0
        if ultimatedespair and movestats and not myorders:
            feelings.append('ultimate despair')
        if despair and movestats and not myorders:
            feelings.append('despair')
        dprint("%s feels: " % myname + str(feelings))
        
        weights = self.unit_weights(mylocation, myunit, feelings)
        
        bestscore = None
        bestmoves = []
        if movestats:
            statlabels = movestats[0].keys()
            statlabels.remove('location')
            statlabels.remove('attack')
            statlabels.remove('attackindex')
            statlabels.remove('attacklocation')
        for action in movestats:
            actionscore = 0
            action['reasons'] = []
            for label in statlabels:
                if action[label] and weights[label]:
                    action['reasons'].append((label, action[label]))
                    actionscore += action[label] * weights[label]
            if action['location'] == mylocation and not action['attack']:
                nomove = action
                if mylocation in self.orders:
                    # this controls how likely units are to follow the
                    # commander's orders... they'll follow the order if
                    # it's not more than twice as bad as doing nothing
                    if not mytype.not_living:
                        if actionscore > 0:
                            actionscore /= 2.0
                        else:
                            actionscore *= 2.0
                    else:
                        # undead units can't disobey
                        actionscore -= 1000
            if actionscore == bestscore:
                bestmoves.append(action)
            elif bestscore is None or actionscore > bestscore:
                bestmoves = [action]
                bestscore = actionscore
            action['score'] = actionscore
        
        newlocation = mylocation
        if bestmoves:
            newbestmoves = []
            for move in bestmoves:
                if mylocation in self.orders:
                    orderverb = self.orders[mylocation][0]
                    orderlocation = self.orders[mylocation][1]
                    if orderverb == 'attack':
                        if orderlocation in self.allunits:
                            unit = self.allunits[orderlocation]
                            name = self.getname(unit)
                        else:
                            name = 'nothing'
                        readableorder = orderverb + ' %s at %s' % \
                            (name, printable(orderlocation))
                    else:
                        readableorder = "move to %s" % \
                            printable(orderlocation)
                    badorder = False
                    if orderverb == 'attack' and not move['attack']:
                        badorder = True
                    elif (orderverb == 'move' \
                    and move['location'] == mylocation \
                    and not orderlocation == mylocation):
                        badorder = True
                    if len(bestmoves) == 1 and badorder:
                        if not myunit.type().not_living:
                            dprint("%s considered the commander's " % myname +\
                                "order to %s carefully, " % readableorder + \
                                "and decided it was stupid.")
                            doover = True
                            newbestmoves = [move]
                            break
                    elif not badorder:
                        newbestmoves.append(move)
                elif move['location'] == mylocation:
                    # calm down horseman... everything will be ok
                    newbestmoves = [move]
                    break
                newbestmoves.append(move)
            if newbestmoves:
                bestmoves = newbestmoves
            bestmove = random.choice(bestmoves)
            if bestmove['location'] == mylocation:
                dprint("%s chose to stay at %s" % \
                    (myname, printable(mylocation)))
                if bestmove['attack']:
                    dprint("%s wants to attack %s" % (myname, \
                        printable(bestmove['attacklocation'])))
                readablereasons = []
                for label, data in bestmove['reasons']:
                    readablereasons.append( \
                        (label, "%g" % data, "%g" % weights[label]))
                dprint("%s thinks: %s" % (myname, readablereasons))
            else:
                dprint("%s chose to move:" % myname)
                dprint("%s was at: %s" % \
                    (myname, printable(mylocation)))
                dprint("%s: score: %s" % (myname, nomove['score']))
                readablereasons = []
                for label, data in nomove['reasons']:
                    readablereasons.append( \
                        (label, "%g" % data, "%g" % weights[label]))
                dprint("%s thought: %s" % (myname, readablereasons))
                dprint("%s went to: %s" % \
                    (myname, printable(bestmove['location'])))
                dprint("%s score: %s" % (myname, bestmove['score']))
                if bestmove['attack']:
                    dprint("%s wants to attack %s" % (myname, \
                        printable(bestmove['attacklocation'])))
                readablereasons = []
                for label, data in bestmove['reasons']:
                    readablereasons.append( \
                        (label, "%g" % data, "%g" % weights[label]))
                dprint("%s thinks: %s" % (myname, readablereasons))
                newlocation = \
                    wesnoth.move_unit(mylocation, bestmove['location'])
                acted = True
            if not self.contactmade and not bestmove['attack']:
                # check if it was after a village. If so, call dibs
                distance, target = self.movestats_targetdistance(mylocation,
                    bestmove['location'], myunit, primarytargets, \
                    returnmaintarget='please')
                if (target and self.gamemap.is_village(target) \
                and not target == bestmove['location'] \
                and not target in self.allunits):
                    dprint("%s calls dibs on the village at %s!" % \
                        (myname, printable(target)))
                    self.villagedibs[target] = myunit.name
            if newlocation in self.villagedibs:
                self.villagedibs.pop(newlocation)
        else:
            if myunit.movement_left:
                wprint("%s couldn't find anywhere to move to!" % myname)
        
        if not myunit.is_valid:  # oh noes, moving killed it! D:
            print '%s screams!' % myname + \
                  ' "YEEAARGHHLLLLLGGHH!!!"'
            return acted, resetturn
        
        if (myunit.can_attack and not doover \
        and bestmoves and bestmove['attack']):
            # ATTACK!!!
            enemylocation = bestmove['attacklocation']
            enemyname = self.getname(self.allunits[enemylocation])
            index = bestmove['attackindex']
            dprint("%s chose attack index: %s" % (myname, index))
            wesnoth.attack_unit(newlocation, enemylocation, index)
            # check if we killed it, and if so, reanalyse orders etc.
            killed = enemylocation not in wesnoth.get_units()
            if (myorders and myorders[0] == 'attack' \
            and myorders[-1] == 'mustkill' and not killed):
                resetturn = True
            if killed:
                resetturn = True
                iprint("%s has slain %s!" % (myname, enemyname))
            elif not enemylocation in self.attackedlocations:
                self.attackedlocations.append(enemylocation)
            acted = True
        elif 'ultimate despair' in feelings or 'despair' in feelings:
            # coward!
            if newlocation == mylocation:
                iprint("...%s gives in to inevitability" % myname)
            else:
                iprint("...%s screams: \"Help! Somebody help me!" % myname + \
                    " I don't want to die!\" ... (coward)")
        
        if doover:
            wprint("%s rebels agains its leader!" % myname)
            self.orders.pop(mylocation)
            # unit should not have moved yet
            acted, resetturn = self.dounit(mylocation, myunit, myname)
        
        return acted, resetturn
        
    def unit_movestats(self, mylocation, myunit, mymoves, primarytargets, \
    secondarytargets):
        '''
        Rate each action the unit can take by a variety of criteria.
        Returns a list of dictionary objects, one for each action
        the unit can perform.
        
        Non-scoring keys are:
            * 'location': the location the unit will move to
            * 'attack': True if the action involves attacking
            * 'attackindex'
            * 'attacklocation'
            * 'village': True if location is on a village
        
        Other keys are the stat names, and values are numbers greater
        than 0.0, with 1.0 denoting a "normal" value. 2.0 means two times
        "more" of the stat, 0.5 means half as much, etc.
        '''
        mytype = myunit.type()
        myname = self.getname(myunit)
        myworth = self.getworth(myunit)
        friendlybysource = wesnoth.get_destinations_by_unit()
        friendlybydest = wesnoth.get_units_by_destination()
        iprint("%s %s is thinking about possible moves..." % \
            (myname, printable(mylocation)))
        
        
        #timea = time.time()
        defaultbysource, defaultbydest = wesnoth.test_move(mylocation)
        #timeb = time.time()
        megadprint("%s fetched default enemy movemap in %s ms" % (myname, \
        #    (timeb - timea) * 1000))
            '?'))
        
        # check what critical locations the enemy can reach ignoring our unit
        enemymapfreedom = self.movestats_mapfreedom(myunit, defaultbydest)
        
        # each move is evaluated according to certain criteria, and each
        # aspect of the move is given a 'value'; how much 'gold' this
        # aspect of this move is worth.
        # see individual functions for more info.
        movestats = []
        
        myorders = None
        myordermoves = [mylocation]
        if mylocation in self.orders:
            dprint("%s is pondering its orders" % myname)
            myorders = self.orders[mylocation]
            orderlocation = myorders[1]
            if myorders[0] == 'attack':
                if orderlocation not in self.allunits:
                    wprint("%s can't follow an order to attack" % myname + \
                        " a unit which isn't there! >:(")
                    myorders = None
                else:
                    enemy = self.allunits[orderlocation]
                    enemyname = self.getname(enemy)
                    if not enemy.is_enemy:
                        wprint("%s refuses to follow its commander's " % \
                            myname + "barbaric order to attack %s!" % \
                            enemyname)
                        myorders = None
                    else:
                        surroundings = \
                            wesnoth.get_adjacent_tiles(orderlocation)
                        for tile in surroundings:
                            if tile in mymoves and not tile == mylocation:
                                # already have mylocation
                                myordermoves.append(tile)
                        if len(myordermoves) == 1:
                            wprint("%s can't get to %s :/" % \
                                (myname, enemyname))
            else:
                if not orderlocation in mymoves:
                    wprint("%s can't get to %s :S" % \
                        (myname, printable(orderlocation)))
                    myorders = None
                else:
                    myordermoves.append(orderlocation)
        
        if myorders:
            mymoves = myordermoves
        
        #lasttime = time.time()
        
        for location in mymoves:
            if location in self.allunits:
                if not self.allunits[location] == myunit:
                    if myorders and myorders[1] == location:
                        unitname = self.getname(self.allunits[location])
                        wprint("%s can't follow its orders"  %  myname + \
                            ", %s is in the way!" % unitname)
                        dprint('%s thinks: "...whatever"' % myname)
                        myorders = None
                    continue
            move = {}
            move['location'] = location
            chancetobehit = myunit.defense_modifier(location) / 100.0
            # Defense value is proportional to the number of turns it would
            # take to kill the unit. Which is inversely proportional to
            # chance to be hit. Worth more if the unit is worth more.
            move['defense'] = 0.5 * myworth / float(chancetobehit)
            move['defense'] -= myworth # call 50% normal
            # defense is actually already part of the danger calculation.
            # so divide it by some random factor here.
            move['defense'] /= 2.0
            move['attack'] = False
            move['attackindex'] = None
            move['attacklocation'] = None
            
            if location in defaultbydest:
                #timea = time.time()
                enemiesbysource, enemiesbydest = \
                    wesnoth.test_move(mylocation, location)
                #timeb = time.time()
                # will throw a NoneType error if something goes wrong.
                megadprint("%s revised movemap for %s in %s ms" % (myname, \
                #    printable(location), (timeb - timea) * 1000))
                    printable(location), '?'))
            else:
                # don't need to recalculate if the move wasn't in their
                # destination map anyway
                enemiesbysource = defaultbysource
                enemiesbydest = defaultbydest
            
            danger, retaliation, deathchance = \
                self.dangercalculator(myunit, location, enemiesbydest)
            
            move['danger'] = self.hptogold(danger, myunit) # for now
            move['retaliation'] = retaliation # already in gold value
            deathvalue = self.getworth(myunit) + \
                self.xptogold(self.xpfromunit(myunit))
            basedeathchance = deathchance
            if deathchance:
                megadprint("%s: move %s has base deathchance %s" % (myname, \
                    printable(location), deathchance))
            move['deathchance'] = deathchance * deathvalue
            
            # base stats for doing nothing
            move['damagetaken'] = 0.0
            move['killchance'] = 0.0
            move['damagegiven'] = 0.0
            
            support, healing, curing = self.movestats_support(myunit, \
                mylocation, location, friendlybydest)
            if (location == mylocation and not danger \
            and myunit.movement_left == myunit.max_movement):
                # rest healing
                healing += 2
            
            move['danger'] /= (support + 1.0)
            move['retaliation'] /= (support + 1.0)
            # retaliation is just there to offset danger, so:
            move['retaliation'] = min(move['retaliation'], move['danger'])
            
            wantedhealing = danger + myunit.max_hitpoints - myunit.hitpoints
            healing = min(healing, wantedhealing)
            move['healing'] = self.hptogold(healing, myunit)
            move['curing'] = self.hptogold(curing, myunit)
            # not sure how to scale support yet. For now it's in sum of
            # base costs of units which can help
            
            move['primarytargetdist'] = self.movestats_targetdistance( \
                mylocation, location, myunit, primarytargets)
            move['secondarytargetdist'] = self.movestats_targetdistance( \
                mylocation, location, myunit, secondarytargets)
            
            if (self.gamemap.is_village(mylocation) \
            and not self.gamemap.is_village(location) \
            and mylocation in defaultbydest \
            and mylocation in enemiesbydest):
                # everything else is NOT YOUR FUCKING PROBLEM, CAPICHE?
                if location == mylocation:
                    hinderance = self.criticallocations[mylocation]
                else:
                    hinderance = 0
            else:
                newenemymapfreedom = self.movestats_mapfreedom( \
                    myunit, enemiesbydest)
                hinderance = enemymapfreedom - newenemymapfreedom
            # less likely to hinder things if you might die
            if deathchance:
                hinderance *= 1.0 - deathchance
            move['enemyhinderance'] = hinderance
            
            move['misc'] = self.movestats_misc(myunit, mylocation, location)
            
            #timenow = time.time()
            #timetaken = timenow - lasttime
            #lasttime = timenow
            megadprint("%s analysed move %s in %s ms" % (myname, \
            #    printable(location), timetaken * 1000))
                printable(location), '?'))
            if (myorders and myorders[0] == 'attack' \
            and location != mylocation):
                # don't consider only moving if the order was to attack
                pass
            else:
                readablemove = []
                for stat, value in move.iteritems():
                    if isinstance(value, float) or isinstance(value, int):
                        readablemove.append((stat, "%g" % value))
                megadprint("results: %s" % readablemove)
                
                
                movestats.append(move)
            
            
            # now do similar for each possible attack
            possibleattacks = []
            for tile in wesnoth.get_adjacent_tiles(location):
                if tile in enemiesbysource: # ignores hidden units
                    if myorders and myorders[0] == 'attack':
                        totile = myorders[1]
                        if not tile == totile:
                            # only considering our orders for now
                            continue
                    for attackindex in xrange(len(myunit.attacks())):
                        possibleattacks.append((tile, attackindex))
            for enemylocation, attackindex in possibleattacks:
                # for the most part, has the same scores as the base move
                attack = move.copy()
                attack['attack'] = True
                attack['attackindex'] = attackindex
                attack['attacklocation'] = enemylocation
                enemy = self.allunits[enemylocation]
                enemyname = self.getname(enemy)
                enemytype = enemy.type()
                enemyworth = self.getworth(enemy)
                myattacktypes = myunit.attacks()
                attacktype = myattacktypes[attackindex]
                
                killchance, deathchance, damagegiven, damagetaken, \
                    maxdamagegiven, maxdamagetaken = self.attackstats( \
                    myunit, enemy, location, enemylocation, attackindex)
                
                if (attacktype.poison \
                and not enemy.poisoned \
                and not enemytype.regenerate):
                    # prefer to poison enemies at 2/3 health
                    poisonvalue = min(16, 7 + enemy.hitpoints)
                    # this was already applied in attackstats so:
                    damagegiven -= poisonvalue
                    poisonvalue /= (1.0 + \
                        abs(enemy.max_hitpoints * 2.0 - \
                        enemy.hitpoints * 3.0))
                    damagegiven += poisonvalue
                if damagetaken < 0:
                    # unit can't drain more hp than it lacks
                    damagetaken = max(damagetaken, \
                        myunit.hitpoints - myunit.max_hitpoints)
                
                if enemylocation in self.attackedlocations:
                    # try to gang up on things more
                    if damagegiven > 0:
                        damagegiven *= 1.1
                if self.gamemap.is_village(enemylocation):
                    # killing units on villages is more important
                    damagegiven *= 1.1
                    if killchance:
                        # killing it prevents a heal
                        damagegiven += 8.0 * killchance
                attack['damagegiven'] = self.hptogold(damagegiven, enemy)
                attack['damagegiven'] += self.xptogold(enemy.type().level)
                attack['damagetaken'] = self.hptogold(damagetaken, myunit)
                attack['damagetaken'] += self.xptogold(myunit.type().level)
                myxptogo = myunit.max_experience - myunit.experience
                enemyxptogo = enemy.max_experience - enemy.experience
                xpfromme = min(self.xpfromunit(myunit), enemyxptogo)
                xpfromenemy = min(self.xpfromunit(enemy), myxptogo)
                killvalue = self.getworth(enemy)
                killvalue += self.xptogold(xpfromenemy)
                if myxptogo <= xpfromenemy:
                    killvalue += self.hptogold(myunit.max_hitpoints - \
                        myunit.hitpoints, myunit)
                attack['killchance'] = killvalue * killchance
                deathvalue = self.getworth(myunit)
                deathvalue += self.xptogold(xpfromme)
                # reverse-calculate original deathchance of move
                if enemyxptogo <= xpfromme:
                    deathvalue += self.hptogold(enemy.max_hitpoints - \
                        enemy.hitpoints, enemy)
                totaldeathchance = \
                    1.0 - (1.0 - deathchance) * (1.0 - basedeathchance)
                megadprint("%s: %s has revised death chance of %s" % \
                    (myname, printable(location), totaldeathchance) + \
                    " due to attack deathchance of %s" % deathchance)
                attack['deathchance'] = totaldeathchance * deathvalue
                # retaliation value will be changed with new death chance
                oldretaliation = \
                    attack['retaliation'] / (1.0 - basedeathchance)
                attack['retaliation'] = min(attack['danger'], \
                    oldretaliation * (1.0 - totaldeathchance))
                
                # TODO: if unit is gamewin, ALL danger is removed on kill
                
                # if you level, it's a full heal!
                if not deathchance:
                    xpfromattack = enemytype.level
                    xptogo = myunit.max_experience - myunit.experience
                    if xpfromattack >= xptogo:
                        fullheal = myunit.max_hitpoints - myunit.hitpoints
                        attack['healing'] = max(attack['healing'], \
                            self.hptogold(fullheal, myunit))
                        attack['damagetaken'] = 0.0
                if killchance:
                    potentialxp = self.xpfromunit(enemy)
                    xptogo = myunit.max_experience - myunit.experience
                    if potentialxp >= xptogo:
                        fullheal = myunit.max_hitpoints - myunit.hitpoints
                        attack['healing'] = max(attack['healing'], \
                            (killchance * self.hptogold(fullheal, myunit)))
                        # damage taken doesn't count if you kill it
                        attack['damagetaken'] *= (1 - killchance)
                    # unit should already be able to weigh healing vs. danger
                if killchance:
                    # if it dies, it's not a threat to our villages.
                    # so kill that damned bat already
                    thisenemybydest = {}
                    for destination in enemiesbysource[enemylocation]:
                        if enemiesbydest[destination] == [enemylocation]:
                            thisenemybydest[destination] = [enemylocation]
                    thisenemymapfreedom = self.movestats_mapfreedom( \
                        myunit, thisenemybydest)
                    thisenemymapfreedom *= killchance
                    attack['enemyhinderance'] += thisenemymapfreedom
                
                #timenow = time.time()
                #timetaken = timenow - lasttime
                #lasttime = timenow
                megadprint("%s analysed attack at %s in %s ms" % (myname, \
                #    printable(location), timetaken * 1000))
                    printable(location), '?'))
                #readablemove = []
                #for stat, value in attack.iteritems():
                #    if isinstance(value, float) or isinstance(value, int):
                #        readablemove.append((stat, "%g" % value))
                #megadprint("results: %s" % readablemove)
                movestats.append(attack)
            
        return movestats
    
    def dangercalculator(self, myunit, location, enemiesbydest, ignore=[]):
        '''
        Returns calculated 'danger', 'retaliation' and 'deathchance' 
        values for myunit at mylocation attacked by the most
        fearsome-looking set of enemies possible.
        
        Only the given enemiesbydestination map is used, and any units
        in 'ignore' are ignored.
        
        danger = total expected damage (in hp)
        retaliation = expected retaliation (in gold value)
        deathchance = approximate chance of dying.
        
        enemiesbydestination is the usual location -> list of locations.
        '''
        # TODO: at some point I would like to add a "maximum danger" value,
        # for the danger if the enemy has total move freedom (i.e. ignoring
        # our other units and ZOC). I don't think there's any way of
        # retrieving the movemaps for 'best possible enemy moves' through
        # the python interface though.
        wesnoth.raise_user_interact()
        myname = self.getname(myunit)
        mytype = myunit.type()
        myworth = self.getworth(myunit)
        myctbh = myunit.defense_modifier(location) / 100.0
        surroundings = wesnoth.get_adjacent_tiles(location)
        attackerbytile = {}
        attackers = []
        totaldamage = 0.0
        totalretaliation = 0.0
        totalmaxdamage = 0.0
        healing = 0.0
        maxsurvivalchance = 1.0
        totalnumattacks = 0.0
        for tile in surroundings:
            if tile not in enemiesbydest:
                continue
            if len(enemiesbydest[tile]) == 1:
                unitlocation = enemiesbydest[tile][0]
                if unitlocation in attackers:
                    continue
                if unitlocation in ignore:
                    continue
                unit = self.allunits[unitlocation]
                attackers.append(unitlocation)
                bestdamage = None
                for attackindex in xrange(len(unit.attacks())):
                    attackstats = self.attackstats(unit, myunit, tile, \
                        location, attackindex, nextturn=True)
                    damage = attackstats[2]
                    # survival chance just makes sure any units which could
                    # kill myunit at it's current hitpoint level are
                    # sufficiently scary.
                    # overall survivalchance is estimated later.
                    if bestdamage is None or damage > bestdamage:
                        bestdamage = damage
                        survivalchance = 1.0 - attackstats[0]
                        retaliation = self.hptogold(attackstats[3], unit)
                        numattacks = unit.attacks()[attackindex].num_attacks
                        maxdamage = attackstats[4]
                totaldamage += bestdamage
                totalretaliation += retaliation
                maxsurvivalchance *= survivalchance
                totalnumattacks += numattacks
                totalmaxdamage += maxdamage
                
                megadprint("%s if at %s could be attacked by %s (worth %s)" % \
                    (myname, printable(location), self.getname(unit), \
                    self.getworth(unit)) + \
                    " for %s likely damage, with %s gold-retaliation" % \
                    (bestdamage, retaliation) + ", max chance of survival: %s" % \
                    survivalchance + ", max damage %s" % maxdamage)
        for tile in surroundings:
            if tile not in enemiesbydest:
                continue
            possibleattackers = []
            tilebestdamage = 0.0
            bestattackunit = None
            for unitlocation in enemiesbydest[tile]:
                if unitlocation in attackers:
                    continue
                if unitlocation in ignore:
                    continue
                unit = self.allunits[unitlocation]
                for attackindex in xrange(len(unit.attacks())):
                    attackstats = self.attackstats(unit, myunit, tile, \
                        location, attackindex, nextturn=True)
                    damage = attackstats[2]
                    # survival chance just makes sure any units which could
                    # kill myunit at it's current hitpoint level are
                    # sufficiently scary.
                    # overall survivalchance is estimated later.
                    if damage > tilebestdamage:
                        tilebestdamage = damage
                        bestattackunit = unitlocation
                        tilesurvivalchance = 1.0 - attackstats[0]
                        retaliation = self.hptogold(attackstats[3], unit)
                        tilenumattacks = unit.attacks()[attackindex].num_attacks
                        tilemaxdamage = attackstats[4]
            if bestattackunit:
                attackers.append(bestattackunit)
                totaldamage += tilebestdamage
                totalmaxdamage += tilemaxdamage
                totalretaliation += retaliation
                maxsurvivalchance *= tilesurvivalchance
                totalnumattacks += tilenumattacks
                unit = self.allunits[bestattackunit]
                megadprint("%s if at %s could be attacked by %s (worth %s)" % \
                    (myname, printable(location), self.getname(unit), \
                    self.getworth(unit)) + \
                    " for %s likely damage, with %s gold-retaliation" % \
                    (tilebestdamage, retaliation) + \
                    ", max chance of survival: %s" % tilesurvivalchance + \
                    ", max damage %s" % tilemaxdamage)
        
        if not totalnumattacks:
            return 0.0, 0.0, 0.0
        if totaldamage <= 0:
            totalretaliation *= maxsurvivalchance
            return totaldamage, totalretaliation, 1.0 - maxsurvivalchance
        # this is so !@#$ing complicated...
        # gonna assume that each potential hit does the same damage
        # thus the probability of a certain number of hits landing
        # is a Binomial Distribution with p = averagecth, q = 1-p,
        # N = total number of attacks. Assuming N is large, this approaches
        # a Normal Distribution around Np with sigma^2 = Npq.
        # the case where N is small will be approximated by including the
        # individual killchances of attacking units, above.
        # see postscript Note 1 for more info.
        N = float(totalnumattacks)
        eachhit = totalmaxdamage / N
        n = hitstokill = myunit.hitpoints / float(eachhit)
        p = averagectbh = totaldamage / float(totalmaxdamage)
        megadprint("final numattacks, eachhit, hitstokill, averagectbh = " + \
            str([N, eachhit, n, p]))
        sigma2 = N * p * (1.0 - p)
        x = (n - N * p) / math.sqrt(2 * sigma2)
        x2 = x**2
        a = 0.147
        erfx = math.sqrt(1 - ( \
            math.exp(-(x2) * ((4/math.pi) + a*x2) / (1 + a*x2))))
        if x >= 0:
            survival = 0.5 * (1 + erfx)
        else:
            survival = 0.5 * (1 - erfx) # booyah
        totalsurvivalchance = maxsurvivalchance * survival
        totalretaliation *= totalsurvivalchance
        
        movedanger = totaldamage # in hp
        moveretaliation = totalretaliation # in gold value
        movedeathchance = 1.0 - totalsurvivalchance
        if movedeathchance < 0.001:
            movedeathchance = 0
        return movedanger, moveretaliation, movedeathchance
    
    def movestats_support(self, myunit, mylocation, location, friendlybydest):
        '''
        Returns:
            * support = combined base cost of allies that can reach the
                  location reduced proportionately if they are damaged
            * healing = potential healing from all sources (in hp)
            * curing = potential curing from all sources (in hp)
        
        Note: 'friendlyunits' can be an iterator.
        '''
        surroundings = wesnoth.get_adjacent_tiles(location)
        support = 0.0
        healing = 0
        curing = 0
        if myunit.type().regenerate or self.gamemap.is_village(location):
            healing = 8
            curing = 8
        for tile in surroundings:
            if tile in self.allunits:
                unit = self.allunits[tile]
                if not unit.is_enemy:
                    if unit.type().curing:
                        healing = 8
                        curing = 8
                    elif unit.type().heals:
                        healing = max(4, healing)
                        # healers stop poison damage... but don't cure it.
                        # plus the healer might decide to move :(
                        curing = max(4, curing)
        # it seems that friendlybydest does not include mylocation for
        # some reason... so take the best score out of all surrounding tiles.
        maxsupport = 0.0
        for tile in surroundings:
            support = 0.0
            if tile not in friendlybydest:
                continue
            for unitlocation in friendlybydest[tile]:
                # will just count the units that can reach this location
                # as well.
                if unitlocation == mylocation:
                    continue
                unit = self.allunits[unitlocation]
                unittype = unit.type()
                unitsupport = unittype.cost
                unitsupport *= unit.hitpoints / float(unit.max_hitpoints)
                if unittype.usage == 'scout':
                    support += unitsupport / 2.0
                if unittype.usage == 'healer':
                    support += unitsupport * 1.5
                else:
                    support += unitsupport
            if support > maxsupport:
                maxsupport = support
        maxsupport /= myunit.type().cost
        return maxsupport, healing, curing
    
    def movestats_mapfreedom(self, myunit, unitsbydest, team=None):
        '''
        Returns an indication of the team's freedom to reach the map's
        defined critical locations. if team isn't supplied, it's assumed
        all enemies are allied.
        
        The return value doesn't mean much unless subtracted from another
        return value with different destinations.
        
        Critical locations are currently:
            * villages and high-defense terrain: from self.analysegamemap
            * around wounded teammates: worth unitworth / 6
            * around dying teammates: worth 2.0 * unitworth / 6
            * around wounded / dying allies: worth half that for teammates
        '''
        if team is None:
            teamisenemy = True
        else:
            teamisenemy = team.is_enemy
        teamcriticallocations = self.criticallocations.copy()
        myworth = self.getworth(myunit)
        for location, unit in self.allunits.iteritems():
            # find the team's wounded  and valuable units, and define the
            # tiles surrounding them as critical locations
            # counts allies wounded units as well, with half weight
            if unit.is_enemy != teamisenemy:
                continue
            if unit == myunit and teamisenemy:
                continue
            unitworth = self.getworth(unit)
            weight = 0.0
            adjacentenemies = 0
            adjacenttiles = wesnoth.get_adjacent_tiles(location)
            for tile in adjacenttiles:
                if tile in self.allunits and self.allunits[tile].is_enemy:
                    adjacentenemies += 1
            if unit.hitpoints * 3 < unit.max_hitpoints:
                if adjacentenemies == 0:
                    weight = unitworth / 3.0
            elif unit.hitpoints * 1.5 < unit.max_hitpoints:
                if adjacentenemies <= 1:
                    weight = unitworth / 6.0
            elif unitworth > myworth:
                # protect higher value units
                weight = unitworth / 6.0
            teamweight = 1.0
            if team is not None:
                if unit.side != team.side:
                    teamweight = 0.5
            weight *= teamweight
            for tile in adjacenttiles:
                if tile in teamcriticallocations:
                    teamcriticallocations[tile] += weight
                else:
                    teamcriticallocations[tile] = weight
        totalscore = 0.0
        for location in unitsbydest:
            # for now, just worth 'weight' if any unit can reach it
            if location in teamcriticallocations:
                if len(unitsbydest[location]):
                    totalscore += (teamcriticallocations[location])
            else:
                # this is for general hinderance of enemy movement,
                # isn't currently worth much
                defense = 1.0 - (myunit.defense_modifier(location) / 100.0)
                totalscore += defense ** 4
        return totalscore
    
    def movestats_targetdistance(self, mylocation, location, myunit, \
    targets, returnmaintarget=False):
        '''
        Calculates the distance to the targets listed in targets[0]
        using the method specified in targets[1].
        
        Method can currently be either 'nearest' or 'average',
        'nearest' counting only the nearest target, and 'average'
        counting all targets and averaging.
        
        It is assumed that being one move closer to a target is worth 4g.
        (derived from the gold difference per turn between an enemy
        holding a village and you holding it)
        '''
        targetlist = targets[0]
        method = targets[-1]
        maintarget = None
        if len(targets) > 2:
            weightlist = targets[1]
        else:
            weightlist = [1.0 for target in targetlist]
        if method == 'nearest' or method == 'nextto':
            originaldist = None
            noweightdist = None
            nextto = 0
            for index in xrange(len(targetlist)):
                # find the nearest target. Distances are divided by target
                # weight, so a target with weight 2.0 that is 20 tiles away
                # will be considered the same distance as one with weight
                # 1.0 that is 10 tiles away.
                dist = mylocation.distance_to(targetlist[index])
                if dist == 1:
                    nextto += 1
                dist /= float(weightlist[index])
                if (originaldist is None \
                or dist < originaldist):
                    # note: if all else is equal, stays with the first
                    # target it finds
                    originaldist = dist
                    noweightdist = dist * float(weightlist[index])
                    maintarget = targetlist[index]
            targetdistance = 0.0
            if originaldist is not None:
                dist = location.distance_to(maintarget)
                # simply -4 for every full move towards it
                targetdistance = 4.0 * (dist - noweightdist) / \
                    float(myunit.max_movement)
            if nextto and method == 'nextto':
                # the more targets you can be next to the better.
                # Used for healers.
                # assumes being next to 1 target is worth 4g for now
                targetdistance = -4 * nextto
        elif method == 'average':
            totaldist = 0.0
            targetcount = 0.0
            for index in xrange(len(targetlist)):
                dist = self.getdistance(myunit, location, targetlist[index])
                dist *= 4.0
                dist /= float(weightlist[index])
                totaldist += dist
                targetcount += 1.0
            if targetcount:
                targetdistance = totaldist / targetcount
            else:
                targetdistance = 0.0
        else:
            wprint("................................................ what?")
            def fail():
                return self
        
        if returnmaintarget:
            return targetdistance, maintarget
        else:
            return targetdistance
    
    def movestats_misc(self, myunit, mylocation, location):
        '''
        For miscellaneous modifiers that don't fit anywhere else.
        '''
        misc = 0.0
        if self.gamemap.is_village(location):
            # bonus for capturing a village:
            # assumes we will hold the village for 4 turns
            villageowned = False
            for team in wesnoth.get_teams():
                if team.owns_village(location):
                    villageowned = True
                    if team.is_enemy:
                        misc += 16.0
            if not villageowned:
                misc += 8.0
            if self.losing: # true if less than 50% of villages are ally-owned
                misc *= 2.0
        else:
            surroundings = wesnoth.get_adjacent_tiles(location)
            # don't stand next to empty villages you idiots!
            for tile in surroundings:
                if self.gamemap.is_village(tile):
                    if (tile in self.allunits \
                    and self.allunits[tile].is_enemy):
                        misc -= 1.0
                    else:
                        misc -= 2.0
                        break
        if myunit.can_recruit and self.gamemap.is_keep(location):
            # better place to stand than nowhere
            # note: distance to this tile is also considered via
            # primarytargetdistance for a leader.
            misc += 2.0
        #if myunit.type().ambush and ... i can't check the terrain type...
        
        return misc

    def reachablefrom(self, fromlocation, unit):
        '''
        Returns a list of all the tiles 'unit' could reach in one turn,
        starting from 'fromlocation'
        '''
        fromx, fromy = fromlocation.x, fromlocation.y
        maxmoves = unit.max_movement
        minx = max(0 ,fromx - maxmoves)
        maxx = min(self.gamemap.x - 1, fromx + maxmoves)
        miny = max(0, fromy - maxmoves)
        maxy = min(self.gamemap.y - 1, fromy + maxmoves)
        reachablelocations = []
        for x in xrange(minx, maxx):
            for y in xrange(miny, maxy):
                tolocation = wesnoth.get_location(x, y)
                # check
                if tolocation.distance_to(fromlocation) > maxmoves:
                    continue
                if unit.find_path(fromlocation, tolocation, maxmoves):
                    reachablelocations.append(tolocation)
        return reachablelocations
    
    def shortest_path(self, unit, fromlocation, tolocation, limit=None):
        '''
        Returns the shortest path from 'fromlocation' to 'tolocation' for
        'unit', or None if a path cannot be found within the specified limit.
        
        'limit' defaults to the remaining movement points of the unit.
        '''
        if limit is None:
            limit = unit.movement_left
        for moves in xrange(1, limit):
            path = unit.find_path(fromlocation, tolocation, moves)
            if path:
                return path
        return None
    
    def unit_feelings(self, mylocation, myunit):
        '''
        Work out how a certain unit is feeling, based on a number of things.
        '''
        mytype = myunit.type()
        hostileunits = wesnoth.get_enemy_units_by_destination()
        feelings = []
        if myunit.poisoned:
            feelings.append('sick')
        hppercent = myunit.hitpoints / float(myunit.max_hitpoints)
        if hppercent == 1 and not myunit.poisoned:
            feelings.append('refreshed')
        elif hppercent < 0.8:
            if hppercent > 0.67:
                feelings.append('bruised')
            elif hppercent > 0.33:
                feelings.append('wounded')
            else:
                feelings.append('dying')
        xptolevel = myunit.max_experience - myunit.experience
        if xptolevel <= 4:
            feelings.append('hyperactive')
        elif xptolevel <= 8:
            feelings.append('psyched')
        elif xptolevel <= 16:
            feelings.append('pumped')
        elif xptolevel > 24:
            feelings.append('jaded')
        danger, retaliation, deathchance = self.dangercalculator(myunit, \
            mylocation, wesnoth.get_enemy_units_by_destination())
        if danger > 20:
            feelings.append('terrified')
        elif danger > 15:
            feelings.append('scared')
        elif danger > 10:
            feelings.append('apprehensive')
        elif danger > 5:
            feelings.append('wary')
        backup = 0
        for location in self.allunits:
            # note: can see allies even when hidden (e.g. by fog)
            unit = self.allunits[location]
            if not unit.is_enemy:
                unitmoves = unit.type().movement
                # see if it can reach us in 1 turn
                # this code shouldn't work :S but it does for now
                if unit.find_path(location, mylocation, unitmoves):
                    backupworth = self.getworth(unit) * \
                        (unit.hitpoints / float(unit.max_hitpoints))
                    backup += backupworth
                # or 2 turns
                elif unit.find_path(location, mylocation, 2 * unitmoves):
                    backupworth = self.getworth(unit) * \
                        (unit.hitpoints / float(unit.max_hitpoints))
                    backup += backupworth / 2.0
        myworth = self.getworth(myunit)
        if backup > (8 * myworth):
            feelings.append('invincible')
        elif backup > (4 * myworth):
            feelings.append('courageous')
        elif backup > (2 * myworth):
            feelings.append('bold')
        if self.losing:
            feelings.append('anxious')
        todmodifier = self.todmodifier(mytype) # maybe should be nextturn
        if todmodifier > 1:
            feelings.append('favourable tod')
        elif todmodifier < 1:
            feelings.append('unfavourable tod')
        if mytype not in self.myrecruittypes:
            # TODO: or if unit is loyal
            feelings.append('unique')
        # TODO: favourable traits
        return feelings
    
    def unit_targets(self, mylocation, myunit):
        '''
        Work out what map locations should be defined as 'primary'
        and 'secondary' targets.
        
        Returns for each of 'primary', 'secondary':
            [list of targets, (optional) list of weights, method specifier]
        list of targets is a list of wesnoth.location objects
        list of weights is an optional list of float objects
        method specifier is one of: 'nearest', 'average'
        '''
        myname = self.getname(myunit)
        primarytargets = []
        secondarytargets = []
        emptyvillages = []
        enemyvillages = []
        capturablevillages = []
        for location in self.villages:
            allyowns = False
            mine = False
            for team in wesnoth.get_teams():
                if team.owns_village(location):
                    if team.is_enemy:
                        enemyvillages.append(location)
                        break
                    elif team.side == self.myteam.side:
                        mine = True
                        break
                    else:
                        allyowns = True
                        break
            if not allyowns and not location in self.allunits:
                emptyvillages.append(location)
                if mine:
                    continue
                elif (location in self.villagedibs \
                and self.villagedibs[location] != myunit.name):
                    # someone else has dibs, don't count it as capturable.
                    # note: a unit might decide to capture it anyway if it
                    # is in range for them this turn.
                    dprint("Removing %s from capturable villages" % \
                        printable(location) + ", %s has dibs." % \
                        self.villagedibs[location])
                    continue
                else:
                    capturablevillages.append(location)
        leaders = self.getleaders()
        enemyleaders = {}
        myleader = None # ally leaders can take care of their fool selves
        for location, leader in leaders:
            if leader.is_enemy:
                enemyleaders[location] = leader
            elif leader.side == self.myteam.side:
                myleader = location
        enemies = {}
        myunits = {}
        woundedallies = {}
        icure = myunit.type().curing
        iheal = myunit.type().heals
        for location, unit in self.allunits.iteritems():
            if unit.is_enemy:
                enemies[location] = unit
            else:
                if unit.side == self.myteam.side:
                    myunits[location] = unit
                if not icure and not iheal:
                    continue
                if self.gamemap.is_village(location):
                    continue
                if (unit.poisoned \
                or unit.hitpoints * 1.0 / unit.max_hitpoints < 0.75):
                    healingdibs = False
                    curingdibs = False
                    for tile in wesnoth.get_adjacent_tiles(location):
                        otherunit = self.allunits.get(tile)
                        if otherunit:
                            if otherunit.type().curing:
                                healingdibs = True
                                curingdibs = True
                                break
                            elif otherunit.type().heals:
                                healingdibs = True
                    if (iheal and not healingdibs) \
                    or (icure and not curingdibs):
                        woundedallies[location] = unit
                    else:
                        dprint("%s won't bother with %s" % (myname, \
                            self.getname(unit)) + ", someone else has dibs")
        usagehint = myunit.type().usage
        if myunit.can_recruit: # leader
            usefulkeeps = []
            weights = []
            for tile in self.keeps:
                if self.keeps[tile]:
                    # assume a keep in the centre of the map is
                    # better than one in the corner, as long as you
                    # are the closest leader to it.
                    mindist = None
                    for leaderlocation, leader in self.getleaders():
                        if not leaderlocation == mylocation:
                            leaderkeepdist = leaderlocation.distance_to(tile)
                            if mindist is None or leaderkeepdist < mindist:
                                mindist = leaderkeepdist
                    if mindist < mylocation.distance_to(tile):
                        # if there's another closer leader, ignore it
                        continue
                    width, height = self.gamemap.x, self.gamemap.y
                    centerx = math.floor(width / 2)
                    centery = math.floor(height / 2)
                    # might be a tile off, but doesn't matter so much
                    center = wesnoth.get_location(centerx, centery)
                    corner = wesnoth.get_location(0, 0)
                    cornerdistance = float(center.distance_to(corner))
                    if not cornerdistance:
                        1 / 0 == "Don't play on maps with 4 tiles or less >:("
                    keepdistance = float(tile.distance_to(center))
                    nummoves = cornerdistance / float(myunit.max_movement)
                    keepweight = 1.0
                    # * 0.5 to be a bit less keen to charge in
                    keepweight += 0.5 * nummoves * \
                        (1.0 - keepdistance / cornerdistance)
                    usablecastletiles = 0
                    for castletile in self.keeps[tile]:
                        unit = self.allunits.get(castletile)
                        if not unit:
                            usablecastletiles += 1
                        elif (unit.side == self.myteam.side \
                        and unit.movement_left == unit.max_movement):
                            usablecastletiles += 1
                    if not usablecastletiles:
                        dprint("%s couldn't use keep at %s" % \
                            (self.getname(myunit), printable(tile)) + \
                            ", all its castle tiles seem to be occupied!")
                        continue
                    keepweight *= usablecastletiles
                    # (number of attached castle tiles)
                    usefulkeeps.append(tile)
                    weights.append(keepweight)
                    megadprint("%s found keep at %s, weight %s" % \
                        (self.getname(myunit), printable(tile), keepweight))
            primarytargets = [usefulkeeps, weights, 'nearest']
            dprint("leader primary targets: %s" % \
                [printable(keep) for keep in usefulkeeps])
            dprint("weights: %s" % weights)
            secondarytargets = [capturablevillages, 'nearest']
        elif usagehint == 'scout':
            primarytargets = [capturablevillages, 'nearest']
            # try to steal villages away from the battle
            # this might need to be optimised a bit, not sure
            secondaryweights = [-1.0] * len(self.allunits)
            secondarytargets = [self.allunits.keys(), secondaryweights, \
                'average']
        elif usagehint == 'healer':
            primarytargets = [woundedallies.keys(), 'nextto']
            secondaryweights = [2.0] * len(capturablevillages) + \
                [1.0] * len(enemies)
            secondarytargets = [capturablevillages + enemies.keys(), \
                secondaryweights, 'nearest']
        else:
            primaryweights = [2.0] * len(capturablevillages) + \
                [1.0] * len(enemies)
            # note: not sure how enemy on village is treated
            # don't really care
            primarytargets = [capturablevillages + enemies.keys(), \
                primaryweights, 'nearest']
            if not self.losing:
                secondarytargets = [enemyleaders.keys(), 'nearest']
            else:
                # if losing, try to spread out and hopefully flank the
                # enemy or capture villages
                secondaryweights = [-1.0] * len(myunits)
                secondarytargets = [myunits.keys(), secondaryweights, \
                    'average']
        return primarytargets, secondarytargets
    
    def unit_weights(self, mylocation, myunit, feelings):
        '''
        Return a dictionary of weights to be applied to the unit's
        movement scores.
        
        A weight of 1.0 means bigger is better,
        -1.0 means smaller is better,
        2.0 means a little bigger is a lot better,
        0.5 means a lot bigger is only a little better,
        etc.
        '''
        mytype = myunit.type()
        myname = self.getname(myunit)
        # weights will give an 'instinctive' idea of which tile is the
        # best tile to move do, taking into account the unit's mood and
        # various preset criteria
        weights = {  # 1 means bigger is better, -1 means smaller is better
            'danger': -1,
            'retaliation': 1,
            'killchance': 1,
            'deathchance': -0.5, # partially accounted for already in danger
            'damagegiven': 1,
            'damagetaken': -1,
            'enemyhinderance': 1,
            'defense': 0.5,  # partially accounted for already in 'danger'
            'primarytargetdist': -1.1,
            'secondarytargetdist': -0.9,
            'healing': 1,
            'curing': 0, # only used if sick
            'misc': 1, # misc whatever
            }
        # weight modifiers:
        # multiply by greater than one means more important
        # multiply by less than one means less important
        if 'refreshed' in feelings:
            pass
        else:
            weights['damagetaken'] *= 1.05
            weights['danger'] *= 1.05
        if 'bruised' in feelings:
            weights['damagegiven'] *= 0.9
            weights['danger'] *= 1.1
            weights['enemyhinderance'] *= 0.9
            weights['killchance'] *= 0.9
            weights['healing'] *= 1.1
            weights['primarytargetdist'] *= 0.9
            weights['secondarytargetdist'] *= 0.9
        elif 'wounded' in feelings:
            weights['damagegiven'] *= 0.9
            weights['danger'] *= 1.2
            weights['enemyhinderance'] *= 0.6
            weights['killchance'] *= 0.9
            weights['healing'] *= 1.2
            weights['primarytargetdist'] *= 0.7
            weights['secondarytargetdist'] *= 0.7
            if 'ultimate despair' in feelings:
                # Poor sucker, you're probably gonna die.
                # You must be angry. Very angry. Why don't you just
                # let it all out? Show THE ENEMY how you feel!
                if not mytype.not_living:
                    iprint("%s refuses to be taken alive!" % myname)
                weights['danger'] = 0.5
                weights['damagegiven'] *= 2.0
                weights['killchance'] *= 1.5
        elif 'dying' in feelings:
            weights['damagegiven'] *= 0.8
            weights['killchance'] *= 0.8
            weights['retaliation'] *= 0.8
            weights['danger'] *= 1.3
            weights['enemyhinderance'] *= 0.3 # don't even bother, mate.
            weights['primarytargetdist'] *= 0.2
            weights['secondarytargetdist'] *= 0.2
            weights['healing'] *= 1.3
            if 'despair' in feelings or 'ultimate despair' in feelings:
                # You can die a coward; or a HERO. Your choice!
                if not mytype.not_living:
                    iprint("%s won't die without a FIGHT!" % myname)
                weights['danger'] = 0.33
                weights['damagegiven'] *= 3.0
                weights['killchance'] *= 3.0
                weights['deathchance'] *= 0.33
        if 'sick' in feelings:
            weights['damagetaken'] *= 1.1
            weights['curing'] = 1.0
        if 'pumped' in feelings:
            weights['killchance'] *= 1.1
            weights['deathchance'] *= 1.1
            weights['danger'] *= 1.1
            weights['damagegiven'] *= 0.9
            weights['damagetaken'] *= 1.1
            weights['enemyhinderance'] *= 0.9
        elif 'psyched' in feelings:
            weights['killchance'] *= 1.2
            weights['deathchance'] *= 1.2
            weights['danger'] *= 1.2
            weights['damagegiven'] *= 0.8
            weights['damagetaken'] *= 1.2
            weights['enemyhinderance'] *= 0.8
        elif 'hyperactive' in feelings:
            weights['killchance'] *= 1.4
            weights['deathchance'] *= 1.5
            weights['danger'] *= 1.5
            weights['damagegiven'] *= 0.6
            weights['damagetaken'] *= 1.5
            weights['enemyhinderance'] *= 0.5 # you're worth more, fool
        elif 'jaded' in feelings:
            weights['killchance'] *= 0.9
            weights['deathchance'] *= 0.9
        if 'wary' in feelings:
            weights['defense'] *= 2.0
        elif 'apprehensive' in feelings:
            weights['defense'] *= 3.0
        elif 'scared' in feelings:
            weights['defense'] *= 4.0
        elif 'terrified' in feelings:
            weights['defense'] *= 5.0
        if 'bold' in feelings:
            weights['damagegiven'] *= 1.1
            weights['danger'] *= 0.9
        elif 'courageous' in feelings:
            weights['damagegiven'] *= 1.2
            weights['danger'] *= 0.8
        elif 'invincible' in feelings:
            weights['damagegiven'] *= 1.3
            weights['deathchance'] *= 0.8
            weights['killchance'] *= 1.2
            weights['danger'] *= 0.7
        if 'favourable tod' in feelings:
            # note: times of day are ALREADY accounted for in damage given
            # etc. This just makes them ESPECIALLY aggressive
            weights['damagegiven'] *= 1.1
            weights['damagetaken'] *= 0.9
        elif 'unfavourable tod' in feelings:
            weights['damagegiven'] *= 0.9
            weights['damagetaken'] *= 1.1
        if 'unique' in feelings:
            # my, aren't you special!
            weights['damagetaken'] *= 1.3
            weights['deathchance'] *= 2.0
        
        if myunit.can_recruit:
            if not self.recruitedthisturn and self.myteam.gold > 0:
                weights['primarytargetdist'] = -(self.myteam.gold / 4.0)
            else:
                # go on - go capture that village
                weights['primarytargetdist'] = -0.25
            weights['danger'] *= 1.2
            weights['deathchance'] = -5.0 # don't die... ever
            weights['enemyhinderance'] *= 0.5
            weights['damagegiven'] *= 0.9
            weights['retaliation'] *= 0.4
        
        if mytype.usage == 'scout':
            weights['primarytargetdist'] *= 3.0 # capturable villages
            weights['secondarytargetdist'] *= 0.6 # average unit distance
            # (average unit distance has two negative weights, so positive)
            weights['defense'] *= 0.8
            weights['danger'] *= 1.1
            weights['enemyhinderance'] *= 0.9
        
        if mytype.usage == 'healer':
            weights['primarytargetdist'] *= 2.0 # nearest wounded ally
            weights['secondarytargetdist'] *= 0.6 # normal primary target
        
        if self.losing: # allies own less than half the villages
            weights['danger'] *= 0.8
            weights['primarytargetdist'] *= 1.5
            weights['secondarytargetdist'] *= 1.5
            weights['enemyhinderance'] *= 2.0
            weights['damagegiven'] *= 1.5
            weights['damagetaken'] *= 0.8
        elif not self.winning:
            # have a decent number of villages, so HOLD ONTO THEM
            weights['enemyhinderance'] *= 2.0
            weights['primarytargetdist'] *= 0.8
            weights['secondarytargetdist'] *= 0.8
        
        if self.winning: # outnumber enemies 2 to 1
            weights['danger'] *= 0.2
            if not mytype.usage == 'scout':
                weights['primarytargetdist'] *= 2.0
                weights['secondarytargetdist'] *= 2.0
            weights['enemyhinderance'] *= 0.3
            weights['damagegiven'] *= 2.0
            weights['damagetaken'] *= 0.5
        
        return weights


TommyAI()

# for profiling: ignore.
#~ def dostuff():
    #~ TommyAI()
#~ import hotshot
#~ prof = hotshot.Profile("hotshot_stats")
#~ prof.runcall(dostuff)
#~ prof.close()

POSTSCRIPT = '''

::: NOTE 1 :::
    
    If 'N' attacks are expected, and each attack has a chance to hit of 'p'
    (assumed to be constant for all attacks considered), then the probability
    'P(n)' of 'n' attacks landing will be given by a BINOMIAL DISTRIBUTION
    (http://mathworld.wolfram.com/BinomialDistribution.html).
    
    If N is large, this tends towards a NORMAL DISTRIBUTION around Np
    (http://mathworld.wolfram.com/NormalDistribution.html) with
    sigma^2 = Np(1-p)
    
    The cumulative distribution function for a normal distribution for 'n'
    attacks (the probability that the number of attacks landing will be <= n)
    is given by:
    
    D(n) = (1/2)(1 + erf( (n - Np)/(sqrt(2Np(1-p))) ))
    where 'erf' is the ERROR FUNCTION (http://mathworld.wolfram.com/Erf.html)
    
    erf can be approximated by:
    erf(x) ~= sqrt(1 - exp(-x^2((4/Pi + ax^2)/(1 + ax^2)))) for x > 0
    erf(-x) = -erf(x) for all x
    where a ~= 0.147
    (www.theorie.physik.uni-muenchen.de/~serge/erf-approx.pdf)
    
    thus giving the formula used to calculate chance of dying from a
    large combined attack.
    
    For a specific set of attacks, this is converted to a chance of survival,
    and multiplied by the chance of surviving each specific unit's attack,
    if it were to attack first.
    
'''
